# -*- coding: utf-8 -*-
"""Created on Mon Jan  8 16:29:18 2018.

@author: erwan

Operations on Curves, where a curve is a (w, I) array tuple

Similar to OriginPro's Simple Curve Math operators


Routine Listing
---------------

- :func:`~radis.misc.curve.curve_distance`
- :func:`~radis.misc.curve.curve_add`
- :func:`~radis.misc.curve.curve_substract`
- :func:`~radis.misc.curve.curve_multiply`
- :func:`~radis.misc.curve.curve_divide`


-------------------------------------------------------------------------------
"""

import warnings

import numpy as np
from scipy.interpolate import interp1d
from scipy.spatial.distance import cdist


def curve_distance(w1, I1, w2, I2, discard_out_of_bounds=True):
    r"""Get a regularized euclidian distance from curve (w1, I1) to curve (w2, I2)

    .. math::

        D(w_1)[i] = \sqrt{ \sum_j (\hat{I_1}[i]  - \hat{I_2}[j] )^2 + (\hat{w_1}[i] - \hat{w_2}[j])^2}

    Where values are normalized as:

    .. math::

        \hat{A} = \\frac{A}{max(A) - min(A)}

    This regularized Euclidian distance minimizes the effect of a small shift in between the two
    curves in case of stiff curves (like a spectrum bandhead can be)

    No interpolation needed neither.

    Distances for out of bounds values is set to nan

    .. warning ::

        This is a distance on both the waverange and the intensity axis.
        It may be used to compensate for a small offset in your experimental
        spectrum (due to wavelength calibration, for instance) but can lead
        to wrong fits easily. Plus, it is very cost-intensive!


    Parameters
    ----------
    w1, I1: array
        range and values for first curve
    w2, I2: array
        range and values for 2nd curve
    discard_out_of_bounds: boolean
        if True, distance for out of bound values is set to nan. Else, it
        will be the distance from the last point.


    Returns
    -------

    w1, Idist: array
        minimal distance from I1 to I2, for each point in (w1, I1)
    """

    norm_w1 = np.max(w1) - np.min(w1)
    norm_w2 = np.max(w2) - np.min(w2)
    norm_I1 = np.max(I1) - np.min(I1)
    norm_I2 = np.max(I2) - np.min(I2)

    dist = cdist(
        np.array((w1 / norm_w1, I1 / norm_I1)).T,
        np.array((w2 / norm_w2, I2 / norm_I2)).T,
    ).min(axis=1)

    # discard out of bound values
    if discard_out_of_bounds:
        b = np.logical_or(w1 < w2.min(), w1 > w2.max())
        dist[b] = np.nan

    return w1, dist


def curve_add(w1, I1, w2, I2, is_sorted=False, kind="linear"):
    """Add curve (w2, I2) from (w1, I1) Linearly interpolates if the two ranges
    dont match. Fills out of bound parameters with nan.

    Similar to OriginPro's "Simple Curve Math Substract"

    .. note::
        for higher accuracy, choose (w2, I2) has the curve with the highest
        resolution


    Parameters
    ----------
    w1, I1: array
        range and values for first curve
    w2, I2: array
        range and values for 2nd curve
    is_sorted: boolean
        (optional) if True, doesnt sort input arrays
    kind: str
        interpolation kind. Default 'linear'. See scipy.interpolate.interp1d


    Returns
    -------

    w1, Iadd: array
        sum ``I1 + I2`` interpolated on the first range ``w1``
    """

    I2_interp = _curve_interpolate(w1, I1, w2, I2, is_sorted=is_sorted, kind=kind)

    return w1, I1 + I2_interp


def curve_substract(w1, I1, w2, I2, is_sorted=False, kind="linear"):
    """Substracts curve (w2, I2) from (w1, I1) Linearly interpolates if the two
    ranges dont match. Fills out of bound parameters with nan.

    Similar to OriginPro's "Simple Curve Math Substract"

    .. note::
        for higher accuracy, choose (w2, I2) has the curve with the highest
        resolution


    Parameters
    ----------
    w1, I1: array
        range and values for first curve
    w2, I2: array
        range and values for 2nd curve
    is_sorted: boolean
        (optional) if True, doesnt sort input arrays
    kind: str
        interpolation kind. Default 'linear'. See scipy.interpolate.interp1d


    Returns
    -------

    w1, Idiff: array
        difference ``I1 - I2`` interpolated on the first range ``w1``
    """

    I2_interp = _curve_interpolate(w1, I1, w2, I2, is_sorted=is_sorted, kind=kind)

    return w1, I1 - I2_interp


def curve_multiply(w1, I1, w2, I2, is_sorted=False, kind="linear"):
    """Multiply curve (w2, I2) with (w1, I1) Linearly interpolates if the two
    ranges dont match. Fills out of bound parameters with nan.

    Similar to OriginPro's "Simple Curve Math Substract"

    .. note::
        for higher accuracy, choose (w2, I2) has the curve with the highest
        resolution


    Parameters
    ----------
    w1, I1: array
        range and values for first curve
    w2, I2: array
        range and values for 2nd curve
    is_sorted: boolean
        (optional) if True, doesnt sort input arrays
    kind: str
        interpolation kind. Default 'linear'. See scipy.interpolate.interp1d


    Returns
    -------
    w1, Iproduct: array
        product ``I1 * I2`` interpolated on the first range ``w1``
    """

    I2_interp = _curve_interpolate(w1, I1, w2, I2, is_sorted=is_sorted, kind=kind)

    return w1, I1 * I2_interp


def curve_divide(w1, I1, w2, I2, is_sorted=False, kind="linear", interpolation=1):
    """Divides curve (w1, I1) by (w2, I2) Linearly interpolates if the two
    ranges dont match. Fills out of bound parameters with nan.

    Similar to OriginPro's "Simple Curve Math Substract"

    .. note::
        for higher accuracy, choose (w2, I2) has the curve with the highest
        resolution


    Parameters
    ----------
    w1, I1: array
        range and values for first curve
    w2, I2: array
        range and values for 2nd curve

    Other Parameters
    ----------------
    is_sorted: boolean
        (optional) if ``True``, assumes that both input arrays are sorted already.
        Default ``False``.
    kind: str
        interpolation kind. Default 'linear'. See scipy.interpolate.interp1d
    interpolation: int, optional
        If 1, interpolate on w1, I1. Else, on w2, I2.
        Default 1

    Returns
    -------
    w1, Idiv: array
        Division ``I1 / I2`` interpolated on the first or second range according to reverseInterpolation
    """
    if interpolation == 1:
        I2_interp = _curve_interpolate(w1, I1, w2, I2, is_sorted=is_sorted, kind=kind)
        output = w1, I1 / I2_interp
    else:
        I1_interp = _curve_interpolate(w2, I2, w1, I1, is_sorted=is_sorted, kind=kind)
        output = w2, I1_interp / I2

    if np.isnan(output[1]).any():
        warnings.warn(
            "Presence of NaN in curve_divide!\nThink about interpolation=2", UserWarning
        )

    return output


def _curve_interpolate(w1, I1, w2, I2, is_sorted=False, kind="linear"):
    """Resample ``I2`` on ``w1``

    For the interpolation to work, this requires to first sort ``w1`` and ``w2``,
    interpolate, then revert to the initial ``w1``

    .. note::
        for higher accuracy, choose (w2, I2) has the curve with the highest
        resolution

    Returns
    -------

    I2_inter: np.array
        ``I2`` interpolated on ``w1``.
    """
    if (abs((w1[1] - w1[0])) < 0.5 * abs((w2[1] - w2[0]))) and (
        abs((w1[-1] - w1[-2])) < 0.5 * abs((w2[-1] - w2[-2]))
    ):
        # w1 looks like the spectra with highest accuracy
        warnings.warn(
            "First spectrum has more resolution than 2nd. Reverse your spectra in interpolation/comparison for a better accuracy"
        )

    # First sort both
    if not is_sorted:  # TODO : add first check with `radis.misc.is_sorted(a) `
        b = np.argsort(w1)
        w1, I1 = w1[b], I1[b]
        # Get reverse sorting array:
        bm1 = np.zeros_like(w1, dtype=np.int64)
        bm1[b] = np.arange(len(w1))

        b = np.argsort(w2)
        w2, I2 = w2[b], I2[b]

    # resample if needed
    if not np.array_equal(w1, w2):
        f = interp1d(w2, I2, kind=kind, bounds_error=False)
        I2 = f(w1)

    # revert to initial array
    if not is_sorted:
        out = I2[bm1]
    else:
        out = I2

    return out



# -*- coding: utf-8 -*-
"""
Created on Mon May  7 17:34:52 2018

@author: erwan
"""

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pytest

import radis
from radis import get_residual, sPlanck
from radis.lbl import SpectrumFactory
from radis.lbl.base import get_waverange
from radis.misc.printer import printm
from radis.misc.progress_bar import ProgressBar
from radis.misc.utils import Default
from radis.test.utils import setup_test_line_databases


def test_linestrength_calculations(*args, **kwargs):
    """Compare and validate linestrength calculations from
    :py:func:`radis.lbl.base.linestrength_from_Einstein`,
    :py:meth:`radis.lbl.base.BaseFactory.calc_linestrength_eq`,
    :py:meth:`radis.lbl.base.BaseFactory.calc_linestrength_non_eq`,

    """

    # Test linestrength calculations
    from radis.io.hitran import hit2df
    from radis.test.utils import getTestFile

    df = hit2df(getTestFile("hitran_co_3iso_2000_2300cm.par"))

    from radis.db.classes import get_molecule_identifier
    from radis.db.molparam import MolParams
    from radis.lbl.base import linestrength_from_Einstein
    from radis.levels.partfunc import PartFuncTIPS

    molpar = MolParams()
    M = get_molecule_identifier("CO")

    Q_arr = df.iso.map({iso: PartFuncTIPS(M, iso).at(296) for iso in df.iso.unique()})
    Ia_arr = df.iso.map(
        {iso: molpar.get(M, iso, "abundance") for iso in df.iso.unique()}
    )

    # Reference linestrengths recomputed from Einstein coefficients match
    # within 0.1%
    S = linestrength_from_Einstein(df.A, df.gp, df.El, Ia_arr, df.wav, Q_arr, 296)
    assert np.allclose(df.int, S, rtol=1e-3, atol=0)

    #%% Now, at 1000 K
    Q_arr = df.iso.map({iso: PartFuncTIPS(M, iso).at(1000) for iso in df.iso.unique()})
    S_1000 = linestrength_from_Einstein(df.A, df.gp, df.El, Ia_arr, df.wav, Q_arr, 1000)

    from radis.lbl.factory import SpectrumFactory

    sf = SpectrumFactory(2000, 3000)
    sf.load_databank(
        path=getTestFile("hitran_co_3iso_2000_2300cm.par"),
        format="hitran",
        parfuncfmt="hapi",
    )

    # TODO : write an example of all the calculation steps in SpectrumFactory
    # sf._reinitialize()  # creates scaled dataframe df1 from df0  # TODO: make a public function.
    # sf.calc_linestrength_eq(1000)

    sf.eq_spectrum(1000)
    # Linestrengths computed from Einstein coefficients by linestrength_from_Einstein
    # match the one from the SpectrumFactory within 0.1%
    assert np.allclose(sf.df1.S, S_1000, rtol=1e-3, atol=0)

    #%% Using non-eq calculations
    S_eq = sf.df1.S.copy()

    sf.non_eq_spectrum(1000, 1000)
    # Linestrengths computed with equilibrium routine (scaling reference) match
    # the ones from the nonequilibrium routine (from Einstein coefficients)
    # within 0.1%
    assert np.allclose(sf.df1.S, S_eq, rtol=1e-3, atol=0)


@pytest.mark.fast
def test_export_populations(plot=True, verbose=True, warnings=True, *args, **kwargs):
    """Check populations calculated in the nonequilibrium module are exported in Spectrum.

    Compare with hardcoded values

    Examples
    --------
    ::
        sf = SpectrumFactory(export_populations="rovib")
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    if verbose:
        printm(">>> test_export_populations")

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4500,
        wavelength_max=4600,
        wstep=0.001,
        cutoff=1e-30,
        path_length=0.1,
        mole_fraction=400e-6,
        isotope=[1],
        medium="vacuum",
        truncation=5,
        export_populations="rovib",
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITRAN-CO-TEST")

    # Populations cannot be calculated at equilibrium (no access to energy levels)
    s = sf.eq_spectrum(300)
    with pytest.raises(ValueError):  # .. we expect this error here!
        pops = s.get_populations("CO", isotope=1, electronic_state="X")

    # Calculate populations using the non-equilibrium module:
    s = sf.non_eq_spectrum(300, 300)
    pops = s.get_populations("CO", isotope=1, electronic_state="X")
    if not "n" in list(pops["rovib"].keys()):
        raise ValueError("Populations not calculated")
    if plot:
        s.plot_populations()

    # Test hardcoded populations
    assert np.isclose(pops["rovib"]["n"].iloc[0], 0.0091853446840826653)
    assert np.isclose(pops["rovib"]["n"].iloc[1], 0.027052543988733215)
    assert np.isclose(pops["rovib"]["n"].iloc[2], 0.04345502115897712)

    # Compare with factory
    # Plot populations:
    with pytest.raises(ValueError):
        sf.plot_populations()  # no isotope given: error expected
    if plot:
        sf.plot_populations("rovib", isotope=1)
        plt.close()  # no need to keep it open, we just tested the function


@pytest.mark.fast
def test_export_rovib_fractions(
    plot=True, verbose=True, warnings=True, *args, **kwargs
):
    """Compare rovib fraction (nu_vib, nl_vib, nu_rot, nl_rot) are calculated
    in the nonequilibrium module

    Examples
    --------
    ::
        sf = SpectrumFactory(export_lines=True, ...)
        sf.misc.export_rovib_fraction = True  # required from 0.9.30
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    if verbose:
        printm(">>> test_export_rovib_fractions")

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4500,
        wavelength_max=4600,
        wstep=0.001,
        cutoff=1e-30,
        path_length=0.1,
        mole_fraction=400e-6,
        isotope=[1],
        medium="vacuum",
        truncation=5,
        export_lines=True,
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITRAN-CO-TEST")
    sf.misc.export_rovib_fraction = True  # required from 0.9.30

    # Calculate populations using the non-equilibrium module:
    s = sf.non_eq_spectrum(300, 300)
    # Test calculated quantities are there
    assert "Qref" in sf.df1.attrs
    assert "Qvib" in sf.df1.attrs
    assert hasattr(sf.df1, "Qrotu")
    assert hasattr(sf.df1, "Qrotl")

    assert hasattr(s.lines, "nu_vib")
    assert hasattr(s.lines, "nl_vib")
    assert hasattr(s.lines, "nu_rot")
    assert hasattr(s.lines, "nl_rot")

    return True


@pytest.mark.fast
def test_populations_CO2_hamiltonian(
    plot=True, verbose=True, warnings=True, *args, **kwargs
):
    """Calculate nonequilibrium modes with the CO2 Hamiltonian

    ..warning::

        as we only use a reduced set of the CO2 effective Hamiltonian (< 3000 cm-1),
        many levels of the Line Database will not appear in the Levels Database.
        We will need to either filter the Line Database beforehand, or run it
        a first time and remove all levels not found.

        This database is obviously not to be used in a Production code!

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    if verbose:
        printm(">>> _test_populations_CO2_hamiltonian")

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavenum_min=2283.7,
        wavenum_max=2285.1,
        wstep=0.001,
        cutoff=1e-30,
        path_length=0.1,
        mole_fraction=400e-6,
        isotope=[1],
        medium="vacuum",
        truncation=5,
        export_populations="rovib",
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITEMP-CO2-HAMIL-TEST", load_energies=True)

    # First run a calculation at equilibrium
    s = sf.eq_spectrum(300)
    s.name = "equilibrium"

    # Now generate vibrational energies for a 2-T model
    # ... Note that this is arbitrary. Lookup Pannier & Dubuet 2020 for more.
    levels = sf.parsum_calc["CO2"][1]["X"].df
    levels["Evib"] = levels.Evib1 + levels.Evib2 + levels.Evib3

    # Calculate populations using the non-equilibrium module:
    # This will crash the first time (see error in docstrings of the function).
    with pytest.raises(AssertionError):
        sf.non_eq_spectrum(300, 300)

    sf.df0.dropna(inplace=True)

    # Retry:
    s_noneq = sf.non_eq_spectrum(300, 300)
    s_noneq.name = "nonequilibrium"

    # Tests:

    #    s.compare_with(s_noneq, plot=plot)
    assert s_noneq.get_power() > 0
    # TODO: implement actual Assertions (right now we're just checking that
    # functions are properly parsed)

    # TODO. Fix below (and move in dedicated test):
    #    s.line_survey()

    return True


# @pytest.mark.needs_connection
def test_optically_thick_limit_1iso(verbose=True, plot=True, *args, **kwargs):
    """Test that we find Planck in the optically thick limit

    In particular, this test will fail if :

    - linestrength are not properly calculated

    - at noneq, linestrength and emission integrals are mixed up

    The test should be run for 1 and several isotopes, because different
    calculations paths are used internally, and this can lead to different
    errors.

    Also, this test is used to run with DEBUG_MODE = True, which will
    check that isotopes and molecule ids are what we expect in all the
    groupby() loops that make the production code very fast.

    Notes
    -----

    switched from large band calculation with [HITRAN-2016]_ to a calculation with
    the embedded [HITEMP-2010]_ fragment (shorter range, but no need to download files)

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    # Force DEBUG_MODE
    DEBUG_MODE = radis.config["DEBUG_MODE"]
    radis.config["DEBUG_MODE"] = True

    try:

        wavenum_min = 2284.2
        wavenum_max = 2284.6

        P = 0.017  # bar
        wstep = 0.001  # cm-1

        Tgas = 1200

        # %% Generate some CO2 emission spectra
        # --------------
        sf = SpectrumFactory(
            wavenum_min=wavenum_min,
            wavenum_max=wavenum_max,
            molecule="CO2",
            mole_fraction=1,
            path_length=0.05,
            cutoff=1e-25,
            truncation=0.5,
            export_populations=False,  #'vib',
            export_lines=False,
            isotope=1,
            wstep=wstep,
            pseudo_continuum_threshold=0,
            pressure=P,
            verbose=False,
        )
        sf.warnings["NegativeEnergiesWarning"] = "ignore"
        sf.load_databank("HITEMP-CO2-TEST")
        pb = ProgressBar(3, active=verbose)
        s_eq = sf.eq_spectrum(Tgas=Tgas, mole_fraction=1, name="Equilibrium")
        pb.update(1)
        s_2T = sf.non_eq_spectrum(
            Tvib=Tgas, Trot=Tgas, mole_fraction=1, name="Noneq (2T)"
        )
        pb.update(2)
        s_4T = sf.non_eq_spectrum(
            Tvib=(Tgas, Tgas, Tgas), Trot=Tgas, mole_fraction=1, name="Noneq (4T)"
        )
        pb.update(3)
        s_plck = sPlanck(
            wavelength_min=2000,  # =wavelength_min,
            wavelength_max=5000,  # =wavelength_max - wstep,   # there is a border effect on last point
            T=Tgas,
        )
        pb.done()

        # %% Post process:
        # MAke optically thick, and compare with Planck

        if plot:
            s_plck.plot(wunit="nm", Iunit="mW/cm2/sr/nm", lw=2)
        for s in [s_eq, s_2T, s_4T]:

            s.rescale_path_length(1e6)

            if plot:
                s.plot(wunit="nm", nfig="same", lw=4)

            if verbose:
                printm(
                    "Residual between opt. thick CO2 spectrum ({0}) and Planck: {1:.2g}".format(
                        s.name,
                        get_residual(s, s_plck, "radiance_noslit", ignore_nan=True),
                    )
                )

            #            assert get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True) < 1e-3
            assert get_residual(s, s_plck, "radiance_noslit", ignore_nan=True) < 0.9e-4
        if plot:
            plt.legend()

        if verbose:
            printm("Tested optically thick limit is Planck (1 isotope): OK")

    finally:
        # Reset DEBUG_MODE
        radis.config["DEBUG_MODE"] = DEBUG_MODE


def test_optically_thick_limit_2iso(verbose=True, plot=True, *args, **kwargs):
    """Test that we find Planck in the optically thick limit

    In particular, this test will fail if :

    - linestrength are not properly calculated

    - at noneq, linestrength and emission integrals are mixed up

    The test should be run for 1 and several isotopes, because different
    calculations paths are used internally, and this can lead to different
    errors.

    Also, this test is used to run with DEBUG_MODE = True, which will
    check that isotopes and molecule ids are what we expect in all the
    groupby() loops that make the production code very fast.

    Notes
    -----

    switched from large band calculation with [HITRAN-2016]_ to a calculation with
    the embedded [HITEMP-2010]_ fragment (shorter range, but no need to download files)

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    # Force DEBUG_MODE
    DEBUG_MODE = radis.config["DEBUG_MODE"]
    radis.config["DEBUG_MODE"] = True

    try:

        wavenum_min = 2284.2
        wavenum_max = 2284.6

        P = 0.017  # bar
        wstep = 0.001  # cm-1

        Tgas = 1200

        # %% Generate some CO2 emission spectra
        # --------------
        sf = SpectrumFactory(
            wavenum_min=wavenum_min,
            wavenum_max=wavenum_max,
            mole_fraction=1,
            path_length=0.05,
            cutoff=1e-25,
            truncation=0.5,
            export_populations=False,  #'vib',
            export_lines=False,
            molecule="CO2",
            isotope=[1, 2],
            wstep=wstep,
            pseudo_continuum_threshold=0,
            pressure=P,
            verbose=False,
        )
        sf.load_databank("HITEMP-CO2-TEST")
        sf.warnings["NegativeEnergiesWarning"] = "ignore"
        #        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        pb = ProgressBar(3, active=verbose)
        s_eq = sf.eq_spectrum(Tgas=Tgas, mole_fraction=1, name="Equilibrium")
        pb.update(1)
        s_2T = sf.non_eq_spectrum(
            Tvib=Tgas, Trot=Tgas, mole_fraction=1, name="Noneq (2T)"
        )
        pb.update(2)
        s_4T = sf.non_eq_spectrum(
            Tvib=(Tgas, Tgas, Tgas), Trot=Tgas, mole_fraction=1, name="Noneq (4T)"
        )
        pb.update(3)
        s_plck = sPlanck(
            wavelength_min=2000,  # =wavelength_min,
            wavelength_max=5000,  # =wavelength_max - wstep,   # there is a border effect on last point
            T=Tgas,
        )
        pb.done()

        # %% Post process:
        # MAke optically thick, and compare with Planck

        for s in [s_eq, s_2T, s_4T]:

            s.rescale_path_length(1e6)

            if plot:

                nfig = "test_opt_thick_limit_2iso {0}".format(s.name)
                plt.figure(nfig).clear()
                s.plot(wunit="nm", nfig=nfig, lw=4)
                s_plck.plot(wunit="nm", nfig=nfig, Iunit="mW/cm2/sr/nm", lw=2)
                plt.legend()

            if verbose:
                printm(
                    "Residual between opt. thick CO2 spectrum ({0}) and Planck: {1:.2g}".format(
                        s.name,
                        get_residual(s, s_plck, "radiance_noslit", ignore_nan=True),
                    )
                )

            #            assert get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True) < 1e-3
            assert get_residual(s, s_plck, "radiance_noslit", ignore_nan=True) < 0.9e-4

        if verbose:
            printm("Tested optically thick limit is Planck (2 isotopes): OK")

    finally:
        # Reset DEBUG_MODE
        radis.config["DEBUG_MODE"] = DEBUG_MODE


def test_get_waverange(*args, **kwargs):

    # 'wunit' is none
    # ...'wmin/wmax' is none
    # ...... wavenumber is passed > should return the same wavenumbers
    assert get_waverange(wavenum_min=10, wavenum_max=20, wunit=Default("cm-1")) == (
        10,
        20,
    )

    # ...... wavelength is passed > should convert and return the wavenumbers
    assert np.isclose(
        get_waverange(
            wavelength_min=1, wavelength_max=2, medium="vacuum", wunit=Default("cm-1")
        ),
        (5000000.0, 10000000.0),
    ).all()

    # ....... passed both wavenumber and wavelength > should throw error
    with pytest.raises(ValueError):
        get_waverange(
            wavenum_min=1000,
            wavenum_max=2000,
            wavelength_min=1,
            wavelength_max=2,
            medium="vacuum",
            wunit=Default("cm-1"),
        )

    # ...... passed neither wavenumber nor wavlength > should throw error
    with pytest.raises(ValueError):
        get_waverange(wunit=Default("cm-1"))

    # ... 'wmin/wmax' are passed as values without accompanying units
    # ...... wavenumber is passed > should throw error due to multiple input wavespace values
    with pytest.raises(ValueError):
        get_waverange(
            wavenum_min=1, wavenum_max=2, wmin=10, wmax=20, wunit=Default("cm-1")
        )

    # ...... wavelength is passed > should throw error due to multiple input wavespace values
    with pytest.raises(ValueError):
        get_waverange(wmin=10, wmax=20, wavelength_min=1, wavelength_max=2)

    # ...... passed both wavenumber and wavelength > should throw error due to multiple input wavespace values
    with pytest.raises(ValueError):
        get_waverange(
            wmin=1,
            wmax=2,
            wavenum_min=10,
            wavenum_max=20,
            wavelength_min=100,
            wavelength_max=200,
            wunit=Default("cm-1"),
        )

    # ...... passed neither wavenumber nor wavelength > should return wavenumber after converting wmin/wmax assuming default units
    assert get_waverange(wmin=10, wmax=20, wunit=Default("cm-1")) == (10.0, 20.0)

    # ... 'wmin/wmax' are passed as values with accompanying units
    # ...... wavenumber is passed > should throw error due to multiple input wavespace values
    with pytest.raises(ValueError):
        get_waverange(
            wavenum_min=10,
            wavenum_max=20,
            wmin=100 * (1 / u.cm),
            wmax=200 * (1 / u.cm),
            wunit=Default("cm-1"),
        )

    # ...... wavelength is passed > should throw error due to multiple input wavespace values
    with pytest.raises(ValueError):
        get_waverange(
            wmin=100 * (1 / u.cm),
            wmax=200 * (1 / u.cm),
            wavelength_min=10,
            wavelength_max=20,
            wunit=Default("cm-1"),
        )

    # ...... passed both wavenumber and wavelength > should throw error due to multiple input wavespace values
    with pytest.raises(ValueError):
        get_waverange(
            wavenum_min=10,
            wavenum_max=20,
            wmin=100 * (1 / u.cm),
            wmax=200 * (1 / u.cm),
            wavelength_min=10,
            wavelength_max=20,
            wunit=Default("cm-1"),
        )

    # ...... passed neither wavenumber nor wavelength > should return wavenumber after converting wmin/wmax from accompanying unit
    assert get_waverange(
        wmin=100 * (1 / u.cm), wmax=200 * (1 / u.cm), wunit=Default("cm-1")
    ) == (100.0, 200.0)

    assert np.isclose(
        get_waverange(
            wmin=1 * u.cm, wmax=2 * u.cm, medium="vacuum", wunit=Default("cm-1")
        ),
        (0.5, 1.0),
    ).all()

    # 'wunit' is not none
    # ... 'wmin/wmax' is none
    # ...... wavenumber is passed > should throw error as wunit can only be passed with wmin/wmax
    with pytest.raises(ValueError):
        get_waverange(wavenum_min=1, wavenum_max=2, wunit="cm")

    # ...... wavelength is passed > should throw error as wunit can only be passed with wmin/wmax
    with pytest.raises(ValueError):
        get_waverange(wavelength_min=1, wavelength_max=2, wunit="cm")

    # ...... passed both wavenumber and wavelength > should throw error as wunit can only be passed with wmin/wmax
    with pytest.raises(ValueError):
        get_waverange(
            wavenum_min=1,
            wavenum_max=2,
            wavelength_min=10,
            wavelength_max=20,
            wunit="cm",
        )

    # ...... passed neither wavenumber nor wavlength > should throw error
    with pytest.raises(ValueError):
        get_waverange(wunit="cm")

    # ... 'wmin/wmax' are passed as values without accompanying units
    # ...... wavenumber is passed > should throw error as only one set of wavespace parameters can be passed
    with pytest.raises(ValueError):
        get_waverange(wmin=1, wmax=2, wavenum_min=10, wavenum_max=20, wunit="cm")

    # ...... wavelength is passed > should throw error as only one set of wavespace parameters can be passed
    with pytest.raises(ValueError):
        get_waverange(wmin=1, wmax=2, wavelength_min=10, wavelength_max=20, wunit="cm")

    # ...... passed both wavenumber and wavelength > should throw error as only one set of wavespace parameters can be passed
    with pytest.raises(ValueError):
        get_waverange(
            wmin=1,
            wmax=2,
            wavenum_min=10,
            wavenum_max=20,
            wavelength_min=100,
            wavelength_max=200,
            wunit="cm",
        )

    # ...... passed neither wavenumber nor wavlength > should return wavenumber after converting wmin/wmax with given wunit
    assert np.isclose(
        get_waverange(wmin=1, wmax=2, wunit="cm"),
        (0.4998637271242577, 0.9997274542485038),
    ).all()

    assert get_waverange(wmin=1, wmax=2, wunit="cm-1") == (1.0, 2.0)

    # ... 'wmin/wmax' are passed as values with accompanying units
    # ...... wavenumber is passed > should throw error as only one set of wavespace parameters can be passed
    with pytest.raises(ValueError):
        get_waverange(
            wmin=1 * u.cm, wmax=2 * u.cm, wavenum_min=10, wavenum_max=20, wunit="cm"
        )

    # ...... wavelength is passed > should throw error as only one set of wavespace parameters can be passed
    with pytest.raises(ValueError):
        get_waverange(
            wmin=1 * u.cm,
            wmax=2 * u.cm,
            wavelength_min=10,
            wavelength_max=20,
            wunit="cm",
        )

    # ...... passed both wavenumber and wavelength > should throw error as only one set of wavespace parameters can be passed
    with pytest.raises(ValueError):
        get_waverange(
            wmin=1 * u.cm,
            wmax=2 * u.cm,
            wavenum_min=1,
            wavenum_max=2,
            wavelength_min=10,
            wavelength_max=20,
            wunit="cm",
        )

    # ...... passed neither wavenumber nor wavlength > should return wavenumber after converting wmin/wmax with given wunit
    assert np.isclose(
        get_waverange(wmin=1 * u.cm, wmax=2 * u.cm, wunit="cm"),
        (0.4998637271242577, 0.9997274542485038),
    ).all()

    assert get_waverange(wmin=1 * (1 / u.cm), wmax=2 * (1 / u.cm), wunit="cm-1") == (
        1.0,
        2.0,
    )
    assert np.isclose(
        get_waverange(wmin=1 * u.cm, wmax=2 * u.cm, wunit="cm"),
        (0.4998637271242577, 0.9997274542485038),
    ).all()
    assert np.isclose(
        get_waverange(
            wavelength_min=1 * u.cm, wavelength_max=2 * u.cm, wunit=Default("cm-1")
        ),
        (0.4998637271242577, 0.9997274542485038),
    ).all()

    # ... passed wmin/wmax with units different from wunit > should throw error due to conflicting units
    with pytest.raises(ValueError):
        get_waverange(wmin=1 * u.cm, wmax=2 * u.cm, wunit="cm-1")


def _run_testcases(verbose=True, plot=True):

    test_linestrength_calculations()
    test_export_populations(plot=plot, verbose=verbose)
    test_export_rovib_fractions(plot=plot, verbose=verbose)
    test_populations_CO2_hamiltonian(plot=plot, verbose=verbose)
    test_optically_thick_limit_1iso(plot=plot, verbose=verbose)
    test_optically_thick_limit_2iso(plot=plot, verbose=verbose)
    test_get_waverange()


if __name__ == "__main__":
    _run_testcases()



"""API for Exomol molecular database

Borrowed from the `Exojax <https://github.com/HajimeKawahara/exojax>`__
code (which you should also have a look at !), by @HajimeKawahara, under MIT License.

"""
import numpy as np
import pandas as pd


def read_def(deff):
    """Exomol IO for a definition file

    Parameters
    ----------
    deff: definition file

    Returns
    -------
    temperature exponent n_Texp
    broadening parameter alpha_ref
    molecular mass
    numinf: nu minimum for trans
    numtag: tag for wavelength range

    Note:
       For some molecules, ExoMol provides multiple trans files. numinf and numtag are the ranges and identifiers for the multiple trans files.


    """

    dat = pd.read_csv(deff, sep="#", names=("VAL", "COMMENT"))
    alpha_ref = None
    # texp = None
    molmasssw = False
    n_Texp = None
    ntransf = 1
    maxnu = 0.0
    quantum_labels = []
    for i, com in enumerate(dat["COMMENT"]):
        if "Default value of Lorentzian half-width" in com:
            alpha_ref = float(dat["VAL"][i])
        elif "Default value of temperature exponent" in com:
            n_Texp = float(dat["VAL"][i])
        elif "Element symbol 2" in com:
            molmasssw = True
        elif "No. of transition files" in com:
            ntransf = int(dat["VAL"][i])
        elif "Maximum wavenumber (in cm-1)" in com:
            maxnu = float(dat["VAL"][i])
            # maxnu=20000.0
        elif molmasssw:
            c = np.unique(dat["VAL"][i].strip(" ").split(" "))
            c = np.array(c, dtype=np.float)
            molmass = np.max(c)
            molmasssw = False

        elif "Lifetime availability" in com:
            lifetime = dat["VAL"][i] == 1
        elif "Lande g-factor availability" in com:
            lande = dat["VAL"][i] == 1

        elif "Quantum label" in com:
            quantum_labels.append(dat["VAL"][i].strip(" "))

        # SOME DEF FILES CONTAINS ERRORS. THESE ARE THE EXCEPTIONS
        if deff.stem == "12C-16O2__UCL-4000":
            ntransf = 20
        if deff.stem == "14N-1H3__CoYuTe":
            maxnu = 20000.0

    if ntransf > 1:
        dnufile = maxnu / ntransf
        numinf = dnufile * np.array(range(ntransf + 1))
        numtag = []
        for i in range(len(numinf) - 1):
            imin = "{:05}".format(int(numinf[i]))
            imax = "{:05}".format(int(numinf[i + 1]))
            numtag.append(imin + "-" + imax)
    else:
        numinf = None
        numtag = ""

    output = {
        "n_Texp": n_Texp,
        "alpha_ref": alpha_ref,
        "molmass": molmass,
        "numinf": numinf,
        "numtag": numtag,
        "quantum_labels": quantum_labels,  # array
        "Landé": lande,  # bool
        "lifetime": lifetime,  # bool
    }
    return output


def read_pf(pff):
    """Exomol IO for partition file

    Note:
        T=temperature
        QT=partition function

    Args:
        pff: partition file
    Returns:
        partition data in pandas DataFrame

    """
    dat = pd.read_csv(pff, sep="\s+", names=("T", "QT"))
    return dat


def read_trans(transf):
    """Exomol IO for a transition file
    Note:
        i_upper=Upper state counting number
        i_lower=Lower state counting number
        A=Einstein coefficient in s-1
        nu_lines=transition wavenumber in cm-1
        See Table 12 in https://arxiv.org/pdf/1603.05890.pdf

    Args:
        transf: transition file
    Returns:
        transition data in pandas DataFrame

    """
    try:
        dat = pd.read_csv(
            transf,
            compression="bz2",
            sep="\s+",
            names=("i_upper", "i_lower", "A", "nu_lines"),
        )
    except:
        dat = pd.read_csv(
            transf, sep="\s+", names=("i_upper", "i_lower", "A", "nu_lines")
        )

    return dat


def read_states(statesf, dic_def):
    """Exomol IO for a state file
    Note:
        i=state counting number
        E=state energy
        g=state degeneracy
        J=total angular momentum
        See Table 11 in https://arxiv.org/pdf/1603.05890.pdf

    Args:
        statesf: state file
        dic_def: Info from def file to read extra quantum numbers
    Returns:
        states data in pandas DataFrame

    """
    # we read first 4 columns for ("i", "E", "g", "J"),
    # skip lifetime, skip Landé g-factor,
    # read quantum numbers
    N = np.size(dic_def["quantum_labels"])
    quantum_labels = dic_def["quantum_labels"]

    if dic_def["Landé"] and dic_def["lifetime"]:
        usecol = np.concatenate((np.arange(4), 5 + 2 + np.arange(N)))
        names = ("i", "E", "g", "J", "Lande", "lifetime") + tuple(quantum_labels)
    elif dic_def["Landé"]:
        usecol = np.concatenate((np.arange(4), 5 + 1 + np.arange(N)))
        names = ("i", "E", "g", "J", "Lande") + tuple(quantum_labels)
    elif dic_def["lifetime"]:
        usecol = np.concatenate((np.arange(4), 5 + 1 + np.arange(N)))
        names = ("i", "E", "g", "J", "lifetime") + tuple(quantum_labels)
    else:  # no lifetime, nor landé according to the def file
        usecol = np.concatenate((np.arange(4), 5 + np.arange(N)))
        names = ("i", "E", "g", "J") + tuple(quantum_labels)

    try:
        dat = pd.read_csv(
            statesf, compression="bz2", sep="\s+", usecols=usecol, names=names
        )
    except:  #!!!TODO What was the expected error?
        dat = pd.read_csv(statesf, sep="\s+", usecols=usecol, names=names)

    return dat


def pickup_gE(states, trans, dic_def, trans_lines=False):
    """extract g_upper (gup), E_lower (elower), and J_lower and J_upper from states DataFrame and insert them to transition DataFrame.

    Args:
       states: states pandas DataFrame
       trans: transition pandas DataFrame
       trans_lines: By default (False) we use nu_lines computed using the state file, i.e. E_upper - E_lower. If trans_nuline=True, we use the nu_lines in the transition file. Note that some trans files do not this info.
       dic_def: Informations about additional quantum labels


    Returns:
       A, nu_lines, elower, gup, jlower, jupper, **quantum_labels

    Note:
       We first convert pandas DataFrame to ndarray. The state counting numbers in states DataFrame is used as indices of the new array for states (newstates). We remove the state count numbers as the column of newstate, i.e. newstates[:,k] k=0: E, 1: g, 2: J. Then, we can directly use the state counting numbers as mask.


    """
    ### Step 1. Essential quantum number for spectra
    ndstates = states.to_numpy()[:, :4]  # the i, E, g, J are in the 4 first columns
    ndtrans = trans.to_numpy()

    iorig = np.array(ndstates[:, 0], dtype=int)
    maxii = int(np.max(iorig) + 1)
    newstates = np.zeros((maxii, np.shape(ndstates)[1] - 1), dtype=float)
    newstates[iorig, :] = ndstates[:, 1:]

    i_upper = np.array(ndtrans[:, 0], dtype=int)
    i_lower = np.array(ndtrans[:, 1], dtype=int)

    # use the state counting numbers (upper and lower) as masking.
    elower = newstates[i_lower, 0]
    eupper = newstates[i_upper, 0]
    gup = newstates[i_upper, 1]
    jlower = np.array(newstates[i_lower, 2], dtype=int)
    jupper = np.array(newstates[i_upper, 2], dtype=int)

    ### Step 2. Extra quantum numbers (e/f parity, vib and rot numbers)
    #!!! Todo: Awfull management of memory, i know! @minou
    n0 = (
        4 + dic_def["Landé"] + dic_def["lifetime"]
    )  # we need to shift if these quantities exist
    ndstates_extra = states.to_numpy()[
        :, n0:
    ]  # the i, E, g, J are in the 4 first columns
    a, b = np.shape(ndstates_extra)
    dico_quantumNumbers = {}
    if b != 0:
        for index, label in enumerate(dic_def["quantum_labels"]):
            if label == "K":  #!!!TODO: allow users to have other quantum numbers
                dico_quantumNumbers["{}_l".format(label)] = ndstates_extra[
                    i_lower, index
                ]
                # dico_quantumNumbers['{}_u'.format(label)] = ndstates_extra[index][i_upper]

    A = ndtrans[:, 2]

    if trans_lines:
        nu_lines = ndtrans[:, 3]
    else:
        nu_lines = eupper - elower

    # See Issue #16
    # import matplotlib.pyplot as plt
    # nu_lines_t=ndtrans[:,3]
    # plt.plot(nu_lines_t-nu_lines,".",alpha=0.03)
    # plt.ylabel("diff nu from trans and state (cm-1)")
    # plt.xlabel("wavenuber (cm-1)")
    # plt.savefig("nudiff.png", bbox_inches="tight", pad_inches=0.0)
    # plt.show()

    return A, nu_lines, elower, gup, jlower, jupper, dico_quantumNumbers


# def pickup_gEslow(states, trans):
#     """Slow version to extract g_upper (gup) and E_lower (elower) from states DataFrame and insert them to transition DataFrame.

#     Note:
#        This code is the (thousands times) slower version of pickup_gE. However, we did not remove this one just because this version is much easier to understand.

#     Args:
#        states: states pandas DataFrame
#        trans: transition pandas DataFrame

#     Returns:
#        A, nu_lines, elower, gup

#     """
#     import tqdm

#     E = states["E"].values
#     g = states["g"].values

#     # insert new columns in transition array
#     trans["gup"] = 0
#     trans["elower"] = 0.0

#     for k, i in tqdm.tqdm(enumerate(states["i"])):
#         # transition upper state
#         mask_upper = trans["i_upper"] == i
#         trans["gup"][mask_upper] = g[k]
#         # transition lower state
#         mask_lower = trans["i_lower"] == i
#         trans["elower"][mask_lower] = E[k]

#     A = trans["A"].to_numpy()
#     nu_lines = trans["nu_lines"].to_numpy()
#     elower = trans["elower"].to_numpy()
#     gup = trans["gup"].to_numpy()
#     return A, nu_lines, elower, gup


# def read_broad(broadf):
#     """Reading braodening file (.broad)

#     Args:
#        broadf: .broad file

#     Return:
#        broadening info in bdat form (pandas), defined by this instance.

#     Note:
#        See Table 16 in https://arxiv.org/pdf/1603.05890.pdf

#     """
#     bdat = pd.read_csv(
#         broadf,
#         sep="\s+",
#         names=(
#             "code",
#             "alpha_ref",
#             "n_Texp",
#             "jlower",
#             "jupper",
#             "kalower",
#             "kclower",
#             "kaupper",
#             "kcupper",
#             "v1lower",
#             "v2lower",
#             "v3lower",
#             "v1upper",
#             "v2upper",
#             "v3upper",
#         ),
#     )
#     return bdat


# def check_bdat(bdat):
#     """cheking codes in .broad

#     Args:
#        bdat: exomol .broad data given by exomolapi.read_broad

#     Returns:
#        code level: None, a0, a1, other codes unavailable currently,

#     """

#     def checkcode(code):
#         cmask = bdat["code"] == code
#         if len(bdat["code"][cmask]) > 0:
#             return True
#         else:
#             return False

#     codelv = None
#     for code in ["a0", "a1"]:
#         if checkcode(code):
#             codelv = code

#     return codelv


# def make_j2b(bdat, alpha_ref_default=0.07, n_Texp_default=0.5, jlower_max=None):
#     """compute j2b (code a0, map from jlower to alpha_ref)

#     Args:
#        bdat: exomol .broad data given by exomolapi.read_broad
#        alpha_ref_default: default value
#        n_Texp_default: default value
#        jlower_max: maximum number of jlower

#     Returns:
#        j2alpha_ref[jlower] provides alpha_ref for jlower
#        j2n_Texp[jlower]  provides nT_exp for jlower

#     """
#     # a0
#     cmask = bdat["code"] == "a0"
#     jlower_arr = np.array(bdat["jlower"][cmask], dtype=int)
#     alpha_ref_arr = np.array(bdat["alpha_ref"][cmask])
#     n_Texp_arr = np.array(bdat["n_Texp"][cmask])

#     if jlower_max is None:
#         Nblower = np.max(jlower_arr) + 1
#     else:
#         Nblower = np.max([jlower_max, np.max(jlower_arr)]) + 1
#     j2alpha_ref = np.ones(Nblower) * alpha_ref_default
#     j2n_Texp = np.ones(Nblower) * n_Texp_default

#     j2alpha_ref[jlower_arr] = alpha_ref_arr
#     j2n_Texp[jlower_arr] = n_Texp_arr

#     Ndef = Nblower - (np.max(jlower_arr) + 1)
#     if Ndef > 0:
#         print(
#             "default broadening parameters are used for ",
#             Ndef,
#             " J lower states in ",
#             Nblower,
#             " states",
#         )

#     return j2alpha_ref, j2n_Texp


# def make_jj2b(bdat, j2alpha_ref_def, j2n_Texp_def, jupper_max=None):
#     """compute jj2b (code a1, map from (jlower, jupper) to alpha_ref and n_Texp)

#     Args:
#        bdat: exomol .broad data given by exomolapi.read_broad
#        j2alpha_ref_def: default value from a0
#        j2n_Texp_def: default value from a0
#        jupper_max: maximum number of jupper

#     Returns:
#        jj2alpha_ref[jlower,jupper] provides alpha_ref for (jlower, jupper)
#        jj2n_Texp[jlower,jupper]  provides nT_exp for (jlower, jupper)

#     Note:
#        The pair of (jlower, jupper) for which broadening parameters are not given, jj2XXX contains None.

#     """
#     # a1
#     cmask = bdat["code"] == "a1"
#     jlower_arr = np.array(bdat["jlower"][cmask], dtype=int)
#     jupper_arr = np.array(bdat["jupper"][cmask], dtype=int)
#     alpha_ref_arr = np.array(bdat["alpha_ref"][cmask])
#     n_Texp_arr = np.array(bdat["n_Texp"][cmask])

#     # Nblower = len(j2alpha_ref_def)

#     if jupper_max is None:
#         Nbupper = np.max(jupper_arr) + 1
#     else:
#         Nbupper = np.max([jupper_max, np.max(jupper_arr)]) + 1

#     jj2alpha_ref = j2alpha_ref_def[:, np.newaxis] * np.ones(Nbupper)
#     jj2n_Texp = j2n_Texp_def[:, np.newaxis] * np.ones(Nbupper)

#     jj2alpha_ref[jlower_arr, jupper_arr] = alpha_ref_arr
#     jj2n_Texp[jlower_arr, jupper_arr] = n_Texp_arr

#     return jj2alpha_ref, jj2n_Texp


if __name__ == "__main__":
    pass
    # import pathlib
    # import sys
    # import time

    # deff = pathlib.Path(
    #     "/home/kawahara/exojax/examples/luhman16/.database/CO2/12C-16O2/UCL-4000/12C-16O2__UCL-4000.def"
    # )
    # n_Texp, alpha_ref, molmass, numinf, numtag = read_def(deff)
    # print(numtag)
    # sys.exit()
    # # various broad file
    # #    broadf="/home/kawahara/exojax/data/broad/12C-16O__H2.broad"
    # broadf = "/home/kawahara/exojax/data/broad/1H2-16O__H2.broad"
    # bdat = read_broad(broadf)
    # codelv = check_bdat(bdat)
    # print(codelv)
    # if codelv == "a0":
    #     j2alpha_ref, j2n_Texp = make_j2b(bdat, jlower_max=100)
    # elif codelv == "a1":
    #     j2alpha_ref, j2n_Texp = make_j2b(bdat, jlower_max=100)
    #     jj2alpha_ref, jj2n_Texp = make_jj2b(bdat, j2alpha_ref, j2n_Texp, jupper_max=100)
    #     print(jj2alpha_ref[1, 2])
    #     print(jj2alpha_ref[1, 15])

    # sys.exit()
    # # broad file
    # broadf = "/home/kawahara/exojax/data/CO/12C-16O/12C-16O__H2.broad"
    # bdat = read_broad(broadf)
    # j2alpha_ref, j2n_Texp = make_j2b(bdat, jlower_max=100)

    # # partition file
    # pff = "/home/kawahara/exojax/data/exomol/CO/12C-16O/Li2015/12C-16O__Li2015.pf"
    # dat = read_pf(pff)

    # check = False
    # if check:
    #     print("Checking compution of Elower and gupper.")
    # statesf = (
    #     "/home/kawahara/exojax/data/exomol/CO/12C-16O/Li2015/12C-16O__Li2015.states.bz2"
    # )
    # states = read_states(statesf)
    # transf = (
    #     "/home/kawahara/exojax/data/exomol/CO/12C-16O/Li2015/12C-16O__Li2015.trans.bz2"
    # )
    # trans = read_trans(transf)

    # ts = time.time()
    # A, nu_lines, elower, gup, jlower, jupper = pickup_gE(states, trans)
    # #    for i in range(0,len(A)):
    # #        print(jlower[i],"-",jupper[i])
    # te = time.time()

    # tsx = time.time()
    # if check:
    #     A_s, nu_lines_s, elower_s, gup_s = pickup_gEslow(states, trans)
    # tex = time.time()
    # print(te - ts, "sec")
    # if check:
    #     print(tex - tsx, "sec for the slow version")
    #     print("CHECKING DIFFERENCES...")
    #     print(np.sum((A_s - A) ** 2))
    #     print(np.sum((nu_lines_s - nu_lines) ** 2))
    #     print(np.sum((elower_s - elower) ** 2))
    #     print(np.sum((gup_s - gup) ** 2))

    # # computing alpha_ref, n_Texp



# -*- coding: utf-8 -*-
"""Print functions, with colors."""

import sys
from io import StringIO

import pandas as pd
from termcolor import colored

# %% Colored Print functions


def printm(*args, **kwargs):
    """Print in magenta.

    Typically use in RADIS for tests
    Could be used for Accuracy/Performance warnings."""

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color="magenta"))


def printg(*args, **kwargs):
    """Print in green.

    Typically use in RADIS for verbose>=2 messages"""

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color="green"))


def printr(*args, **kwargs):
    """Print in red.

    Typically use in RADIS to print errors that are corrected on-the-fly
    by the code"""

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color="red"))


def _capture_print(*args, **kwargs):
    """Emulate print option but get output in a ```str``` variable instead of
    stdout."""

    # Change the output to capture the string instead of sending it to the console
    old_stdout = sys.stdout
    sys.stdout = newstdout = StringIO()  # capture all print
    try:
        print(*args, **kwargs)
        string = newstdout.getvalue()  # Get string output
    except:
        raise
    finally:
        sys.stdout = old_stdout

    # discard line return character (will be added again when we print ``string`` )
    return string[:-1]


# %% Print in pandas


def print_full(x):
    """Print full Pandas series.

    From https://stackoverflow.com/questions/19124601/pretty-print-an-entire-pandas-series-dataframe
    """
    pd.set_option("display.max_rows", len(x))
    print(x)
    pd.reset_option("display.max_rows")


def get_print_full(x):
    """Same as print_full, but returns string."""

    # Change the output to capture the string instead of sending it to the console
    old_stdout = sys.stdout
    sys.stdout = newstdout = StringIO()  # capture all print
    try:
        pd.set_option("display.max_rows", len(x))
        print(x)
        string = newstdout.getvalue()  # Get string output
    except:
        raise
    finally:
        sys.stdout = old_stdout
        pd.reset_option("display.max_rows")

    return string[:-1]



# -*- coding: utf-8 -*-
"""
Test that line survey works

-------------------------------------------------------------------------------

"""

import os
from os.path import exists

import pytest

from radis import SpectrumFactory
from radis.misc.printer import printm
from radis.test.utils import getTestFile, setup_test_line_databases
from radis.tools.database import load_spec


@pytest.mark.fast
def test_line_survey(verbose=True, plot=False, warnings=True, *args, **kwargs):
    """Test line survey"""

    _temp_file = "radis_test_line_survey.html"
    if exists(_temp_file):
        os.remove(_temp_file)

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.line_survey(overlay="abscoeff", writefile=_temp_file)

    assert exists(_temp_file)

    with open(_temp_file) as f:
        d = f.read()
    assert "Linestrength" in d
    assert "Wavenumber" in d

    if verbose:
        print("test_line_survey: html file was correctly generated")

    if not plot:
        # clean after use
        os.remove(_temp_file)

    return True


@pytest.mark.fast
def test_line_survey_CO2(verbose=True, plot=True, warnings=True, *args, **kwargs):

    setup_test_line_databases()

    pl = SpectrumFactory(
        wavenum_min=2380,
        wavenum_max=2400,
        #                         wavelength_min=4170,
        #                         wavelength_max=4200,
        mole_fraction=400e-6,
        path_length=100,  # cm
        cutoff=1e-30,
        isotope=[1],
        export_lines=True,
        save_memory=True,
    )  # 0.2)
    pl.warnings["MissingSelfBroadeningWarning"] = "ignore"
    pl.load_databank("HITRAN-CO2-TEST")
    s = pl.eq_spectrum(Tgas=1500)
    s.apply_slit(0.5)
    if plot:
        s.line_survey(overlay="transmittance", barwidth=0.01)

    if verbose:
        printm("no boolean defined for test_line_survey")
    return True  # test not defined (just testing methods work)


def _run_testcases(plot=True, verbose=True, *args, **kwargs):

    # Show media line_shift
    test_line_survey(plot=plot, verbose=verbose, *args, **kwargs)
    test_line_survey_CO2(plot=plot, verbose=verbose, *args, **kwargs)

    return True


if __name__ == "__main__":

    print("Testing line survey functions:", _run_testcases(plot=True))



# -*- coding: utf-8 -*-
"""Test parsers.

Notes
-----

Runs tests for radis/io so that they can be accessed by pytest (and hopefully
the CI test suite)

Examples
--------

Run all tests::

        pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)::

        pytest -m fast

-------------------------------------------------------------------------------
"""

import os
from os.path import getmtime
from warnings import warn

import numpy as np
import pytest

from radis.io.cdsd import cdsd2df
from radis.io.hitran import hit2df
from radis.misc.warning import IrrelevantFileWarning
from radis.test.utils import getTestFile, setup_test_line_databases


@pytest.mark.fast
def test_hitran_names_match(verbose=True, warnings=True, *args, **kwargs):
    """Compare that HITRAN species defined in :mod:`radis.io.hitran` match
    the nomenclature dictionary : :py:data:`radis.io.hitran.trans`.

    This should be ensured by developers when adding new species.
    """
    from radis.db.classes import (
        HITRAN_CLASS1,
        HITRAN_CLASS2,
        HITRAN_CLASS3,
        HITRAN_CLASS4,
        HITRAN_CLASS5,
        HITRAN_CLASS6,
        HITRAN_CLASS7,
        HITRAN_CLASS8,
        HITRAN_CLASS9,
        HITRAN_CLASS10,
        HITRAN_MOLECULES,
    )
    from radis.misc.basics import compare_lists

    all_hitran = (
        HITRAN_CLASS1
        + HITRAN_CLASS2
        + HITRAN_CLASS3
        + HITRAN_CLASS4
        + HITRAN_CLASS5
        + HITRAN_CLASS6
        + HITRAN_CLASS7
        + HITRAN_CLASS8
        + HITRAN_CLASS9
        + HITRAN_CLASS10
    )
    all_hitran = list(set(all_hitran))

    # All species in HITRAN groups should be valid HITRAN_MOLECULES names
    for m in all_hitran:
        if not m in HITRAN_MOLECULES:
            raise ValueError(
                "{0} is defined in HITRAN groups but has no HITRAN id".format(m)
            )

    # Species in 'HITRAN_MOLECULES' should be classified in groups, else nonequilibrium
    # calculations are not possible.
    if warnings and all_hitran != HITRAN_MOLECULES:
        warn(
            "Difference between HITRAN groups (left) and HITRAN id "
            + "dictionary (right). Some HITRAN species are not classified in "
            + "groups. Nonequilibrium calculations wont be possible for these!:\n"
            + "{0}".format(
                compare_lists(
                    all_hitran, HITRAN_MOLECULES, verbose=False, return_string=True
                )[1]
            )
        )
    return


@pytest.mark.fast
def test_local_hitran_co(verbose=True, warnings=True, **kwargs):
    """Analyse some default files to make sure everything still works."""

    # 1. Load
    df = hit2df(getTestFile("hitran_CO_fragment.par"), cache="regen")
    if verbose:
        print("Read hitran_CO_fragment.par")
        print("---------------------------")
        print(df.head())

    # 2. Test
    assert list(df.loc[0, ["vu", "vl"]]) == [4, 4]
    assert df.dtypes["vu"] == np.int64
    assert df.dtypes["vl"] == np.int64

    return True


def test_local_hitran_co2(verbose=True, warnings=True, **kwargs):

    # 1. Load
    df = hit2df(getTestFile("hitran_CO2_fragment.par"), cache="regen")
    if verbose:
        print("Read hitran_CO2_fragment.par")
        print("----------------------------")
        print(df.head())

    # 2. Test
    assert list(
        df.loc[0, ["v1u", "v2u", "l2u", "v3u", "v1l", "v2l", "l2l", "v3l"]]
    ) == [4, 0, 0, 0, 0, 0, 0, 1]
    assert df.dtypes["v1l"] == np.int64
    assert df.dtypes["v3u"] == np.int64

    return True


def test_local_hitran_h2o(verbose=True, warnings=True, **kwargs):

    # 1. Load
    df = hit2df(getTestFile("hitran_2016_H2O_2iso_2000_2100cm.par"), cache="regen")
    if verbose:
        print("Read hitran_2016_H2O_2iso_2000_2100cm.par")
        print("-----------------------------------------")
        print(df.head())

    # 2. Test
    assert list(df.loc[0, ["v1u", "v2u", "v3u", "v1l", "v2l", "v3l"]]) == [
        0,
        2,
        0,
        0,
        1,
        0,
    ]

    assert df.loc[26, "ju"] == 5  # in .par : line 27, column 99-100
    assert df.loc[27, "ju"] == 18  # in .par : line 28, column 99-100

    assert df.dtypes["v1l"] == np.int64
    assert df.dtypes["v3u"] == np.int64

    assert df.dtypes["ju"] == np.int64
    assert df.dtypes["Kau"] == np.int64
    assert df.dtypes["Kcu"] == np.int64
    assert df.dtypes["jl"] == np.int64
    assert df.dtypes["Kal"] == np.int64
    assert df.dtypes["Kcl"] == np.int64

    return True


def test_local_hitemp_file(verbose=True, warnings=True, **kwargs):
    """Analyse some default files to make sure everything still works."""

    # 1. Load
    df = cdsd2df(
        getTestFile("cdsd_hitemp_09_header.txt"), cache="regen", drop_non_numeric=True
    )
    if verbose:
        print(df.head())

    # 2. Tests
    assert df.wav[3] == 2250.00096
    # make sure P Q R is correctly replaced by drop_non_numeric:
    assert "branch" in df
    assert df["branch"].iloc[0] == 0  # Q
    assert df["branch"].iloc[1] == 1  # R

    return True


def test_irrelevant_file_loading(*args, **kwargs):
    """check that irrelevant files (irrelevant wavenumber) are not loaded"""

    # For cdsd-hitemp files :

    # Ensures file is not empty
    df = cdsd2df(getTestFile("cdsd_hitemp_09_header.txt"), cache="regen")
    assert len(df) > 0

    # Now test nothing is loaded if asking for outside the wavenumber range
    with pytest.raises(IrrelevantFileWarning):
        df = cdsd2df(getTestFile("cdsd_hitemp_09_header.txt"), load_wavenum_min=100000)
    with pytest.raises(IrrelevantFileWarning):
        df = cdsd2df(getTestFile("cdsd_hitemp_09_header.txt"), load_wavenum_max=0.5)

    # For HITRAN files :

    # Ensures file is not empty
    df = hit2df(getTestFile("hitran_2016_H2O_2iso_2000_2100cm.par"), cache="regen")
    assert len(df) > 0

    # Now test nothing is loaded if asking for outside the wavenumber range
    with pytest.raises(IrrelevantFileWarning):
        df = hit2df(
            getTestFile("hitran_2016_H2O_2iso_2000_2100cm.par"), load_wavenum_min=100000
        )
    with pytest.raises(IrrelevantFileWarning):
        df = hit2df(
            getTestFile("hitran_2016_H2O_2iso_2000_2100cm.par"), load_wavenum_max=0.5
        )


def _run_example(verbose=False):
    from radis import SpectrumFactory

    setup_test_line_databases(
        verbose=verbose
    )  # add HITEMP-CO2-TEST in ~/radis.json if not there
    sf = SpectrumFactory(
        wavelength_min=4165,
        wavelength_max=4200,
        path_length=0.1,
        pressure=20,
        molecule="CO2",
        isotope="1",
        cutoff=1e-25,  # cm/molecule
        truncation=5,  # cm-1
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITRAN-CO2-TEST")  # this database must be defined in ~/radis.json


def test_cache_regeneration(verbose=True, warnings=True, **kwargs):
    """Checks that if a line database is manually updated (last edited time changes),
    its cache file will be regenerated automatically
    """

    # Initialisation
    # --------------
    # Run case : we generate a first cache file
    _run_example(verbose=verbose)
    # get time when line database file was last modified
    file_last_modification = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par")
    )
    # get time when cache file was last modified
    cache_last_modification = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.h5")
    )

    # To be sure, re run-example and make sure the cache file was not regenerated.
    _run_example(verbose=verbose)
    assert cache_last_modification == getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.h5")
    )

    # Test
    # ----
    # Now we fake the manual editing of the database.
    # change the time when line database was last modified :
    stinfo = os.stat(getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par"))
    access_time = stinfo.st_atime
    os.utime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par"),
        (file_last_modification + 1, access_time + 1),
    )

    file_last_modification_again = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par")
    )

    assert file_last_modification_again > file_last_modification

    # Run case : this should re-generated the cache file
    _run_example(verbose=verbose)

    cache_last_modification_again = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.h5")
    )

    assert cache_last_modification_again > cache_last_modification


def _run_testcases(verbose=True, *args, **kwargs):

    test_hitran_names_match(verbose=verbose, *args, **kwargs)
    test_local_hitran_co(verbose=verbose, *args, **kwargs)
    test_local_hitran_co2(verbose=verbose, *args, **kwargs)
    test_local_hitran_h2o(verbose=verbose, *args, **kwargs)
    test_local_hitemp_file(verbose=verbose, *args, **kwargs)
    test_irrelevant_file_loading()
    test_cache_regeneration(verbose=verbose, *args, **kwargs)
    return True


if __name__ == "__main__":
    print("Testing test_hitran_cdsd.py: ", _run_testcases(verbose=True))
    test_cache_regeneration(verbose=3)



# -*- coding: utf-8 -*-
"""Tools to deal with HDF5 cache files HDF5 cache files are used to cache
Energy Database files, and Line Database files, and yield a much faster access
time.

Routine Listing
---------------


- :func:`~radis.misc.cache_files.check_cache_file`
- :func:`~radis.misc.cache_files.check_not_deprecated`
- :func:`~radis.misc.cache_files.save_to_hdf`

See Also
--------

:py:func:`~radis.io.hitran.hit2df`,
:py:func:`~radis.io.cdsd.cdsd2df`

-------------------------------------------------------------------------------
"""
# TODO: start using .feather format. Faster than HDF5 for pandas Dataframe,
# and can be multithreaded.
# see:
# https://gist.github.com/gansanay/4514ec731da1a40d8811a2b3c313f836
# and pd.read_feather(file, nthreads=3)


# Note: don't import unicode_literals because it breaks the df.to_hdf of
# save_to_hdf because of a stupid unicode/str error in Python 2.7
import os
from os.path import exists, splitext
from warnings import warn

import pandas as pd
from packaging.version import parse

import radis

try:
    from .hdf5 import HDF5Manager
except ImportError:  # if file is ran as a module
    from radis.io.hdf5 import HDF5Manager
from radis.misc.basics import compare_dict, is_float
from radis.misc.printer import printm, printr
from radis.misc.warning import DeprecatedFileWarning, IrrelevantFileWarning

"""str: forces to regenerate cache files that were created in a previous version"""

# Just make sure LAST_BACKWARD_COMPATIBLE_VERSION is valid
assert radis.__version__ >= radis.config["OLDEST_COMPATIBLE_VERSION"]

# Utils


def load_h5_cache_file(
    cachefile,
    use_cached,
    valid_if_metadata_is={},
    relevant_if_metadata_above={},
    relevant_if_metadata_below={},
    current_version="",
    last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
    verbose=True,
    engine="pytables",
):
    """Function to load a h5 cache file.

    Parameters
    ----------
    cachefile: str
        cache file path
    use_cached: str
        use cache file if value is not ``False``:

        - if ``True``, use (and generate if doesnt exist) cache file.
        - if ``'regen'``, delete cache file (if exists) so it is regenerated
        - if ``'force'``, use cache file and raises an error if it doesnt exist

        if using the cache file, check if the file is deprecated. If it
        is deprecated, regenerate the file unless ``'force'`` was used
        (in that case, raise an error)
    valid_if_metadata_is: dict
        values are compared to cache file attributes. If they dont match,
        the file is considered deprecated. See ``use_cached`` to know
        how to handle deprecated files

        .. note::
            if the file has extra attributes they are not compared
    current_version: str
        version is compared to cache file version (part of attributes).
        If current version is superior, a simple warning is triggered.
    last_compatible_version: str
        if file version is inferior to this, file is considered deprecated.
        See ``use_cached`` to know how to handle deprecated files.
    relevant_if_metadata_above, relevant_if_metadata_below : dict
        values are compared to cache file attributes. If they don't match,
        the function returns a :py:class:`~radis.misc.warning.IrrelevantFileWarning`.
        For instance, load a line database file, only if it contains wavenumbers
        between 2300 and 2500 cm-1 ::

                load_h5_cache_file(..., relevant_if_metadata_above={'wav':2300};
                relevant_if_metadata_below={'wav':2500})

        Note that in such an example, the file data is not read. Only the
        file metadata is. If the metadata does not contain the key (e.g.: ``'wav'``)
        a :py:class:`~radis.misc.warning.DeprecatedFileWarning` is raised.

    Returns
    -------
    df: pandas DataFrame, or None
        None if no cache file was found, or if it was deleted
    """
    # TODO @dev: refactor to use HDF5Manager more
    # 1. know if we have to load the file
    if not use_cached:
        return None
    elif use_cached == "regen" and exists(cachefile):
        os.remove(cachefile)
        if verbose:
            printm("Deleted h5 cache file : {0}".format(cachefile))
        return None

    # 2. check the file is here
    if not exists(cachefile):
        if use_cached == "force":
            raise ValueError("Cache file {0} doesnt exist".format(cachefile))
        else:
            return None  # File doesn't exist. It's okay.

    # 3. read file attributes to know if it's deprecated
    try:
        check_not_deprecated(
            cachefile,
            metadata_is=valid_if_metadata_is,
            metadata_keys_contain=list(relevant_if_metadata_above.keys())
            + list(relevant_if_metadata_below.keys()),
            current_version=current_version,
            last_compatible_version=last_compatible_version,
            engine=engine,
        )
    # ... if deprecated, raise an error only if 'force'
    except DeprecatedFileWarning as err:
        if use_cached == "force":
            raise err
        else:
            if verbose:
                printr(
                    "File {0} deprecated:\n{1}\nDeleting it!".format(
                        cachefile, str(err)
                    )
                )
            os.remove(cachefile)
            return None

    # 4. File is not not deprecated: read the the extremum wavenumbers.    raise
    if relevant_if_metadata_above is not None or relevant_if_metadata_below is not None:
        try:
            check_relevancy(
                cachefile,
                relevant_if_metadata_above,
                relevant_if_metadata_below,
                engine=engine,
                verbose=verbose,
            )
        # ... if irrelevant, raise an error only if 'force'
        except IrrelevantFileWarning as err:
            if verbose >= 2:
                from radis.misc.printer import printg

                printg("Database file {0} irrelevant and not loaded".format(cachefile))
            raise err

    # 5. File is relevant: read the content.
    df = None
    if verbose >= 2:
        printm("Reading cache file ({0})".format(cachefile))
    try:
        # Load file :
        manager = HDF5Manager(engine)
        df = manager.load(cachefile, key="df")

    except KeyError as err:  # An error happened during file reading.
        # Fail safe by deleting cache file (unless we explicitely wanted it
        # with 'force')
        if use_cached == "force":
            raise
        else:
            if verbose:
                printr(
                    "An error happened during cache file reading "
                    + "{0}:\n{1}\n".format(cachefile, str(err))
                    + "Deleting cache file to regenerate it"
                )
            os.remove(cachefile)
            df = None

    return df


def get_cache_file(fcache, engine="pytables", verbose=True):
    """Load HDF5 cache file.

    Parameters
    ----------
    fcache: str
        file name

    Other Parameters
    ----------------
    verbose: bool
        If >=2, also warns if non numeric values are present (it would make
        calculations slower)

    Notes
    -----

    we could start using FEATHER format instead. See notes in cache_files.py
    """
    # TODO @dev: refactor : we probably don't need this function (only used in astroquery)

    # Load file
    manager = HDF5Manager(engine)
    df = manager.load(fcache, key="df")

    # Check file
    # ... 'object' columns slow everything down (not fixed format strings!)
    if verbose >= 2:
        _warn_if_object_columns(df, fcache)

    return df


def check_cache_file(
    fcache, use_cached=True, expected_metadata={}, verbose=True, engine="auto"
):
    """Quick function that check status of cache file generated by RADIS:

    The function first checks the existence of ``fcache``. What is does depends
    on the value of ``use_cached``:

    - if ``True``, check it exists and remove the file if it is not valid.
    - if ``'regen'``, delete cache file even if valid, to regenerate it later.
    - if ``'force'``, raise an error if file doesnt exist.

    Then look if it is deprecated (we just look at the attributes, the file
    is never fully read). Deprecation is done by :py:func:`~radis.io.cache_files.check_not_deprecated`
    comparing the ``metadata=`` content.

    - if deprecated, deletes it to regenerate later unless 'force' was used

    Parameters
    ----------
    fcache: str
        cache file name
    use_cached: ``True``, ``False``, ``'force'``, ``'regen'``
        see notes above. Default ``True``.
    expected_metadata: dict
        attributes to check
    verbose: boolean
        print stuff

    Returns
    -------

    None
        whether the file was valid or not (and was removed).
        Raises a :py:class:`~radis.misc.warning.DeprecatedFileWarning` for
        un unvalid file in mode ``'force'``. The error can be caught by the
        parent function.

    See Also
    --------

    :py:func:`~radis.io.cache_files.check_not_deprecated`
    """

    # Test existence of file:
    if not use_cached:
        return  # we dont want a cache file, no need to test it
    elif use_cached == "regen":
        if exists(fcache):
            os.remove(fcache)
            if verbose:
                print(("Deleted h5 cache file : {0}".format(fcache)))
    elif use_cached == "force":
        if not exists(fcache):
            raise ValueError("Cache file {0} doesnt exist".format(fcache))
    else:  # use_cached == True
        pass  # just use the file as is

    # If file is still here, test if it is deprecated:
    # (we just read the attributes, the file is never fully read)
    if exists(fcache):
        if verbose:
            print(("Using cache file: {0}".format(fcache)))
        try:
            check_not_deprecated(
                fcache,
                metadata_is=expected_metadata,
                current_version=radis.__version__,
                last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
                engine=engine,
            )
        except DeprecatedFileWarning as err:
            if use_cached == "force":
                raise
            else:  # delete file to regenerate it in the end of the script
                if verbose:
                    printr(
                        "File {0} deprecated:\n{1}\nDeleting it!".format(
                            fcache, str(err)
                        )
                    )
                os.remove(fcache)

    return


def check_not_deprecated(
    file,
    metadata_is={},
    metadata_keys_contain=[],
    current_version=None,
    last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
    engine="auto",
):
    """Make sure cache file is not deprecated: checks that ``metadata`` is the
    same, and that the version under which the file was generated is valid.

    Parameters
    ----------
    file: str
        a `` .h5`` cache file for Energy Levels
    metadata_is: dict
        expected values for these variables in the file metadata. If the values dont match,
        a :py:func:`~radis.misc.warning.DeprecatedFileWarning` error is raised.
        If the file metadata contains additional keys/values, no error is raised.
    metadata_keys_contain: list
        expected list of variables in the file metadata. If the keys are not there,
        a :py:func:`~radis.misc.warning.DeprecatedFileWarning` error is raised.

    Other Parameters
    ----------------
    current_version: str, or ``None``
        current version number. If the file was generated in a previous version
        a warning is raised. If ``None``, current version is read from
        :data:`radis.__version__`.
    last_backward_compatible_version: str
        If the file was generated in a non-compatible version, an error is raised.
        (useful parameter to force regeneration of certain cache files after a
         breaking change in a new version)
    engine: ``'h5py'``, ``'pytables'``, ``'vaex'``, ``'auto'``
        which HDF5 library to use. If ``'auto'``, try to guess. Note: ``'vaex'``
        uses ``'h5py'`` compatible HDF5.
    """
    if engine == "auto":
        engine = HDF5Manager.guess_engine(file)

    # Get metadata :
    manager = HDF5Manager(engine)
    file_metadata = manager.read_metadata(file)

    # Raise an error if version is not found
    try:
        file_version = file_metadata.pop("version")
    except KeyError:
        raise DeprecatedFileWarning(
            "File {0} is deprecated : ".format(file)
            + "RADIS version missing in metadata. Delete it to regenerate it on next run"
        )

    # Get current version
    if current_version is None:
        current_version = radis.__version__

    # If file version is anterior to a major change
    # ... Update here versions afterwhich Deprecated cache file is not safe
    # ... (example: a key name was changed)
    if parse(file_version) < parse(last_compatible_version):
        raise DeprecatedFileWarning(
            "File {0} has been generated in a deprecated ".format(file)
            + "version ({0}). Oldest compatible version is {1}. ".format(
                file_version, last_compatible_version
            )
            + "Delete the file to regenerate it on next run"
        )

    # If file version is outdated: Warning, but no error
    if parse(current_version) > parse(file_version):
        warn(
            DeprecationWarning(
                "File {0} has been generated in ".format(file)
                + "a deprecated version ({0}) compared to current ({1})".format(
                    file_version, current_version
                )
                + ". Delete it to regenerate it on next run"
            )
        )
        out = False
    elif parse(current_version) == parse(file_version):
        out = True
    else:
        raise ValueError(
            "Cache file ({0}) generated with a future version ({1} > {2})? ".format(
                file, file_version, current_version
            )
            + "Do you own a DeLorean? Delete the file manually if you understand what happened"
        )

    # Make sure metadata keys are there:
    for k in metadata_keys_contain:
        if k not in file_metadata:
            raise DeprecatedFileWarning(
                "Metadata in file {0} doesn't contain the expected key `{1}`. ".format(
                    file, k
                )
            )

    # Compare metadata values
    # ignore additional keys in the file attributes.
    metadata_is = _h5_compatible(metadata_is)
    file_metadata = {k: v for k, v in file_metadata.items() if k in metadata_is}
    out, compare_string = compare_dict(
        metadata_is,
        file_metadata,
        verbose=False,
        return_string=True,
        df1_str="Expected",
        df2_str="Got",
    )
    if out != 1:
        raise DeprecatedFileWarning(
            "Metadata in file {0} dont match ".format(file)
            + "expected values. See comparison below:"
            + "\n\tExpected\tFile\n{0}".format(compare_string)
        )

    return out


def _h5_compatible(a_dict):
    """Make dictionary ``a_dict`` h5 compatible."""
    out = {}
    for k, v in a_dict.items():
        if v is None:
            continue  # dont store None
        elif is_float(v):
            out[k] = v
        else:
            out[k] = str(v)  # convert to str
    return out


def check_relevancy(
    file,
    relevant_if_metadata_above,
    relevant_if_metadata_below,
    verbose=True,
    key="df",
    engine="auto",
):
    """Make sure cache file is relevant.

    Use case: checks that  wavenumber min and
    wavenumber max in ``metadata`` are relevant for the specified spectral
    range.

    Parameters
    ----------
    file: str
        a `` .h5``  line database cache file
    load_only_wavenum_above, relevant_if_metadata_below: dict
        only load the cached file if the metadata values are above/below
        the specific values for each key.
    relevant_if_metadata_above, relevant_if_metadata_below: dict
        file is relevant if the file metadata value for each key of the dictionary
        is above/below the value in the dictionary

    Other Parameters
    ----------------
    key: str
        dataset key in storer.

    Examples
    --------
    You want to compute a spectrum in between 2300 and 2500 cm-1. A line database
    file is relevant only if its metadata says that ``'wavenum_max' > 2300``
    and ``'wavenum_min'`` < 2500 cm-1.

        check_relevancy('path/to/file', relevant_if_metadata_above={'wavenum_max':2300},
                        relevant_if_metadata_below={'wavenum_min':2500})

        the specified value.

    """
    if engine == "auto":
        engine = HDF5Manager.guess_engine(file)

    # Get metadata :
    manager = HDF5Manager(engine)
    file_metadata = manager.read_metadata(file, key=key)

    for k, v in relevant_if_metadata_above.items():
        # Note : check_not_deprecated already tested the existence of each key so we are safe
        if file_metadata[k] < v:
            raise IrrelevantFileWarning(
                "Database file {0} irrelevant: {1}={2} [file metadata] < {3} [expected], not loaded".format(
                    file, k, file_metadata[k], v
                )
            )
    for k, v in relevant_if_metadata_below.items():
        # Note : check_not_deprecated already tested the existence of each key so we are safe
        if file_metadata[k] > v:
            raise IrrelevantFileWarning(
                "Database file {0} irrelevant ({1}={2} [file metadata] > {3} [expected]), not loaded".format(
                    file, k, file_metadata[k], v
                )
            )


def _warn_if_object_columns(df, fname):
    """'object' columns slow everything down (not fixed format strings!)"""
    objects = [k for k, v in df.dtypes.items() if v == object]
    if len(objects) > 0:
        warn(
            "Dataframe in {0} contains `object` format columns: {1}. ".format(
                fname, objects
            )
            + "Operations will be slower. Try to convert them to numeric."
        )


def save_to_hdf(
    df, fname, metadata, version=None, key="df", overwrite=True, verbose=True
):
    """Save energy levels or lines to HDF5 file. Add metadata and version.

    Parameters
    ----------
    df: a pandas DataFrame
        data will be stored in the key ``'df'``
    fname: str
        ``.h5`` file where to store.
    metadata: dict
         dictionary of values that were used to generate the DataFrame. Metadata
         will be asked again on file load to ensure it hasnt changed. ``None``
         values are not stored.
    version: str, or ``None``
        file version. If ``None``, the current :data:`radis.__version__` is used.
        On file loading, a warning will be raised if the current version is
        posterior, or an error if the file version is set to be uncompatible.
    key: str
        dataset name. Default ``'df'``
    overwrite: boolean
        if ``True``, overwrites file. Else, raise an error if it exists.

    Other Parameters
    ----------------
    verbose: bool
        If >=2, also warns if non numeric values are present (it would make
        calculations slower)

    Notes
    -----
    ``None`` values are not stored
    """
    # Check file
    assert fname.endswith(".h5")
    assert "version" not in metadata
    # ... 'object' columns slow everything down (not fixed format strings!)
    if verbose >= 2:
        _warn_if_object_columns(df, fname)

    # Update metadata format
    metadata = _h5_compatible(metadata)

    # Overwrite file
    if exists(fname) and not overwrite:
        raise ValueError("File exist: {0}".format(fname))

    # start by exporting dataframe
    df.to_hdf(fname, key, format="fixed", mode="a", complevel=9, complib="blosc")

    # Add metadata

    # ... add RADIS version
    if version is None:
        version = radis.__version__
    metadata.update({"version": version})

    with pd.HDFStore(fname, mode="a") as store:
        store.get_storer(key).attrs.metadata = metadata

    if verbose >= 3:
        print("... saved {0} with metadata: {1}".format(fname, metadata))


def filter_metadata(arguments, discard_variables=["self", "verbose"]):
    """Filter arguments (created with  ``locals()`` at the beginning of the
    script) to extract metadata.

    Metadata is stored as attributes in the cached
    file:

    - remove variables in ``discard_variables``
    - remove variables that start with ``'_'``
    - remove varibles whose value is ``None``

    Parameters
    ----------
    arguments: dict
        list of local variables. For instance::

            arguments = locals()
    discard_variables: list of str
        variable names to discard

    Returns
    -------
    metadata: dict
        a (new) dictionary built from arguments by removing ``discard_variables``
        and variables starting with ``'_'``

    Examples
    --------
    How to get only function argument::

        def some_function(*args):
            metadata = locals()     # stores only function arguments because it's the first line

            ...

            metadata = filter_metadata(metadata)
            save_to_hdf(df, fname, metadata=metadata)

            ...
    """
    # TODO: adding metadata this way is not explicit & hard to maintain, should be removed.
    metadata = {k: v for (k, v) in arguments.items() if not k.startswith("_")}
    metadata = {k: v for (k, v) in metadata.items() if k not in discard_variables}
    metadata = {k: v for (k, v) in metadata.items() if v is not None}

    return metadata


def cache_file_name(fname):
    """Return the corresponding cache file name for fname."""
    return splitext(fname)[0] + ".h5"


if __name__ == "__main__":
    import pytest

    # Run relevant tests (here:  the ones with 'cache' in their name)
    printm("Testing cache files:", pytest.main(["../test/io/", "-k", "cache"]))
    printm("Testing cache files:", pytest.main(["../test/lbl/", "-k", "cache"]))



# -*- coding: utf-8 -*-
"""Created on Thu May 28 14:47:36 2015.

@author: Erwan
"""


import json
import os
from collections import OrderedDict
from os.path import abspath

from radis.misc.basics import is_number


def getFile(*relpath):
    r"""Converts the relative path of a database file in a the full path.
    Used by processing script not to worry about where the database is stored

    Examples of use::

        radis.db.getFile('CN','CN_Violet_vib.dat')
        radis.db.getFile('CN\CN_Violet_vib.dat')
        radis.db.getFile('CN\\CN_Violet_vib.dat')

    """

    #    return os.path.join(os.path.dirname(__file__), *relpath)
    from radis.misc.utils import getProjectRoot

    return os.path.join(getProjectRoot(), "db", *relpath)


def check_molecule_data_structure(fname, verbose=True):
    """Check that ``fname`` has a valid JSON structure for molecular data.

    Parameters
    ----------

    fname: str
        molecular data JSON file


    Notes
    -----

    Order in the json doesnt matter, however, the order of the
    electronic_level_names matters: the index of all levels should
    match the `index` key of these states.
    """

    with open(fname) as f:
        try:
            db = json.load(f)  # , object_pairs_hook=OrderedDict)
        except json.JSONDecodeError as err:
            raise json.JSONDecodeError(
                "Error reading '{0}' (line {2} col {3}): \n{1}".format(
                    fname, err.msg, err.lineno, err.colno
                ),
                err.doc,
                err.pos,
            ) from err

    for molecule, db_molec in db.items():
        # ... Check number of isotopes is correct
        isotope_names = db_molec["isotopes_names"]
        isotopes = db_molec["isotopes"]
        if len(isotopes) != len(isotope_names):
            raise ValueError(
                "In molecule {0}: isotope names ".format(molecule)
                + "({0}) dont match the number of isotopes ({1})".format(
                    isotope_names, list(isotopes.keys())
                )
            )

        # ... Check number of electronic states is correct
        for isotope, db_iso in db_molec["isotopes"].items():
            elec_states_names = db_iso["electronic_levels_names"]
            elec_states = db_iso["electronic_level"]
            if len(elec_states_names) != len(elec_states):
                raise ValueError(
                    "In molecule {0}, isotope {1}: electronic ".format(
                        molecule, isotope
                    )
                    + "levels names ({0}) dont match the number of levels ({1})".format(
                        elec_states_names, list(elec_states.keys())
                    )
                )

            # ... Check they are properly ordered
            for state, db_state in elec_states.items():
                if elec_states_names.index(db_state["name"]) + 1 != db_state["index"]:
                    # ... + 1 because Python index start at 0 and FORTRAN at 1 unless allocated with 0:N
                    raise ValueError(
                        "In molecule {0}, isotope {1}: index of electronic ".format(
                            molecule, isotope
                        )
                        + "state {0} ({1}): {2} does not match the list of states: {3}. ".format(
                            state,
                            db_state["name"],
                            db_state["index"],
                            elec_states_names,
                        )
                    )

    if verbose:
        print("Structure of {0} looks correct".format(fname))


def parse_doi(rovib_coeffs):
    if "#doi" in rovib_coeffs:
        doi = rovib_coeffs["#doi"]
    elif "#DOI" in rovib_coeffs:
        doi = rovib_coeffs["#DOI"]
    else:
        doi = None

    return doi


def _get_rovib_coefficients(
    molecule, isotope, electronic_state, jsonfile, remove_trailing_cm1=True
):
    """Returns all rovib coefficients for ``molecule``, ``isotope``,
    ``electronic_state`` by parsing a JSON file of molecule data.

    Parameters
    ----------

    molecule: str
        molecule name

    isotope: int
        isotope number

    electronic_state: str
        electronic state name

    jsonfile: str, or ``default``
        path to json file.

    Other Parameters
    ----------------

    remove_trailing_cm1: bool
        if ``True``, remove the trailing '_cm-1' in the spectroscopic coefficient name,
        if defined. Ex::

            ('we1_cm-1', 1333.93)
            >>> ('we1', 1333.93)

    See Also
    --------

    :py:func:`~radis.db.utils.get_dunham_coefficients`,
    :py:func:`~radis.db.utils.get_herzberg_coefficients`,
    """

    check_molecule_data_structure(jsonfile, verbose=False)

    with open(jsonfile) as f:
        try:
            db = json.load(f, object_pairs_hook=OrderedDict)
        except json.JSONDecodeError as err:
            raise json.JSONDecodeError(
                "Error reading '{0}' (line {2} col {3}): \n{1}".format(
                    jsonfile, err.msg, err.lineno, err.colno
                ),
                err.doc,
                err.pos,
            ) from err

    # Get Dunham coefficients in 001 state (X)
    elec_state_names = db[molecule]["isotopes"][str(isotope)]["electronic_levels_names"]

    try:
        elec_state_index = elec_state_names.index(electronic_state)
    except ValueError:
        raise ValueError(
            "{0} not in the electronic state list for {1}(iso={2}): {3}".format(
                electronic_state, molecule, isotope, elec_state_names
            )
        )
    else:
        elec_state_index = "{:03d}".format(elec_state_index + 1)  # 1 based index

    rovib_coeffs = db[molecule]["isotopes"][str(isotope)]["electronic_level"][
        elec_state_index
    ]

    if remove_trailing_cm1:

        def remove_cm1(coef):
            """Remove the trailing '_cm-1' in the spectroscopic coefficient
            name, if defined."""
            if coef.endswith("_cm-1"):
                coef = coef[:-5]
            return coef

        rovib_coeffs = {remove_cm1(k): v for (k, v) in rovib_coeffs.items()}

    return rovib_coeffs


def get_default_jsonfile(molecule):
    r"""Return full path of default jsonfile for spectroscopic constants for a
    molecule.

    These are stored in:

            radis\db\[molecule]\*.json

    and defined in radis/config.json
    """

    from radis import config

    name = config["spectroscopic_constants"][molecule]

    return abspath(getFile("{0}/{1}".format(molecule, name)))


def get_dunham_coefficients(
    molecule, isotope, electronic_state, jsonfile="default", return_doi=False
):
    r"""Returns Dunham coefficients ``Yij`` for ``molecule``, ``isotope``, ``electronic_state``
    by parsing a JSON file of molecule data.

    Dunham coefficients are identified as starting with ``Y`` (i.e. ``alpha_e``
    won't be recognized)

    Parameters
    ----------
    molecule: str
        molecule name
    isotope: int
        isotope number
    electronic_state: str
        electronic state name
    jsonfile: str, or ``default``
        path to json file. If ``default``, the default JSON file
        defined in radis/config.json is used. See :py:func:`~radis.db.utils.get_default_jsonfile`

    Other Parameters
    ----------------
    return_doi: bool
        if True, returns the value of key ``"#doi"`` or ``"#DOI"`` if it exists,
        and None if it doesn't.

    Returns
    -------
    dict: coeffs
    str: doi, if ``return_doi``

    See Also
    --------

    :py:func:`~radis.db.utils.get_herzberg_coefficients`

    """

    if jsonfile == "default":
        jsonfile = get_default_jsonfile(molecule)

    rovib_coeffs = _get_rovib_coefficients(
        molecule, isotope, electronic_state, jsonfile=jsonfile
    )

    # Only get Dunham coeffs, i.e, these that start with Y
    dunham_coeffs = {k: v for (k, v) in rovib_coeffs.items() if k.startswith("Y")}

    if len(dunham_coeffs) == 0:
        raise ValueError(
            "No Dunham coefficients (Yij) found for {0} {2}(iso={1}) in {3}".format(
                molecule, isotope, electronic_state, jsonfile
            )
            + "\n\nGot: {0}".format(rovib_coeffs.keys())
        )

    if return_doi:
        return dunham_coeffs, parse_doi(rovib_coeffs)
    else:
        return dunham_coeffs


def get_herzberg_coefficients(
    molecule, isotope, electronic_state, jsonfile="default", return_doi=False
):
    r"""Returns spectroscopic coefficients with Herzberg conventions for
    ``molecule``, ``isotope``, ``electronic_state``  by parsing a JSON file of molecule data.

    Herzberg coefficients are the usual:

    .. math::

        \\omega_e, \\alpha_e, B_e, D_e, etc.

    The extensive list of parameters considered as Herzberg coefficients is found in
    :py:data:`~radis.db.conventions.herzberg_coefficients`

    Parameters
    ----------
    molecule: str
        molecule name
    isotope: int
        isotope number
    electronic_state: str
        electronic state name
    jsonfile: str, or ``default``
        path to json file. If ``default``, the ``molecules_data`` JSON file
        in the ``radis.db`` database is used. See :py:func:`~radis.db.utils.get_default_jsonfile`

    Other Parameters
    ----------------
    return_doi: bool
    if True, returns the value of key ``"#doi"`` or ``"#DOI"`` if it exists,
    and None if it doesn't.

    Returns
    -------
    dict: coeffs
    str: doi, if ``return_doi``

    See Also
    --------
    :py:data:`~radis.db.conventions.herzberg_coefficients`,
    :py:func:`~radis.db.utils.get_dunham_coefficients`

    """

    from radis.db.conventions import herzberg_coefficients

    if jsonfile == "default":
        jsonfile = get_default_jsonfile(molecule)

    rovib_coeffs = _get_rovib_coefficients(
        molecule, isotope, electronic_state, jsonfile=jsonfile
    )

    #    # Only get Herzberg coeffs, i.e, these that are defined in :py:data:`~radis.db.conventions.herzberg_coefficients`

    herzberg_coeffs = {
        k: v
        for (k, v) in rovib_coeffs.items()
        if ignore_trailing_number(k) in herzberg_coefficients
    }

    if len(herzberg_coeffs) == 0:
        raise ValueError(
            "No Herzberg spectroscopic coefficients (we, Be, etc.) found for {0} {2}(iso={1}) in {3}".format(
                molecule, isotope, electronic_state, jsonfile
            )
            + "\n\nGot: {0}".format(rovib_coeffs.keys())
        )

    if return_doi:
        return herzberg_coeffs, parse_doi(rovib_coeffs)
    else:
        return herzberg_coeffs


def ignore_trailing_number(coef):
    """Used so that ``wexe1`` matches ``wexe`` as a well defined Herzberg
    coefficient."""

    if is_number(coef[-1]):
        coef = coef[:-1]
    return coef



# -*- coding: utf-8 -*-
"""Created on Fri Jul  6 13:52:04 2018.

@author: erwan
"""
# TODO refactor : rename this file as hitran_utils.py

import numpy as np
import pandas as pd


def parse_hitran_file(fname, columns, count=-1):
    """Parse a file under HITRAN ``par`` format. Parsing is done in binary
    format with :py:func:`numpy.fromfile` so it's as fast as possible.

    Parameters
    ----------
    fname: str
        filename
    columns: dict
        list of columns and their format

    Other Parameters
    ----------------
    count: int
        number of lines to read. If ``-1`` reads all file.

    Returns
    -------
    df: pandas DataFrame
        dataframe with lines

    See Also
    --------
    Used in :py:func:`~radis.io.hitran.hit2df` and :py:func:`~radis.io.cdsd.cdsd2df`
    """

    # To be faster, we read file totally in bytes mode with fromfiles. But that
    # requires to properly decode the line return character:

    # We cannot simply infer it from the OS :
    # problem arise when file was written in an OS and read in another OS (for instance,
    # line return characters are not converted when read from .egg files). Here
    # we read the first line and infer the line return character for it
    data = _read_hitran_file(
        fname, columns, count=1, linereturnformat="a2"
    )  # 'a2' allocates space to get \n or \n\r linereturn formats
    linereturnformat = _get_linereturnformat(data, columns, fname)

    # Now re-read with correct line return character
    data = _read_hitran_file(fname, columns, count, linereturnformat)

    # Return a Pandas dataframe
    return _ndarray2df(data, columns, linereturnformat)


def _get_linereturnformat(data, columns, fname=""):
    """
    Get line return character & format (size).

    Notes
    -----

    We cannot simply infer it from the OS :
    problem arise when file was written in an OS and read in another OS (for instance,
    line return characters are not converted when read from .egg files). Here
    we read the first line and infer the line return character for it
    """
    # fname just for the error message

    # get format (size) of line return
    from radis.misc.basics import to_str

    linereturn = to_str(data[0][-1])
    if to_str("\r\n") in linereturn:
        linereturnformat = "a2"
    elif to_str("\n") in linereturn or to_str("\r") in linereturn:
        linereturnformat = "a1"
    else:
        raise ValueError(
            "Unknown Line return format: {0}. Check that your file {1} has the HITRAN format. First line : {2}".format(
                linereturn, fname, data[0]
            )
        )

    return linereturnformat


def _ndarray2df(data, columns, linereturnformat):
    """"""

    # ... Cast to new type
    # This requires to recast all the data already read, but is still the fastest
    # method I found to read a file directly (for performance benchmark see
    # CDSD-HITEMP parser)
    newtype = [c[0] if (c[1] == str) else c[1] for c in columns.values()]
    dtype = list(zip(list(columns.keys()), newtype)) + [
        ("_linereturn", linereturnformat)
    ]
    data = _cast_to_dtype(data, dtype)

    # %% Create dataframe
    df = pd.DataFrame(data.tolist(), columns=list(columns.keys()) + ["_linereturn"])

    # Delete dummy column than handled the line return character
    del df["_linereturn"]

    # Update format
    for k, c in columns.items():
        if c[1] == str:
            df[k] = df[k].str.decode("utf-8")

    # Strip whitespaces around PQR columns (due to 2 columns jumped)
    if "branch" in df:  # (only in CDSD)
        df["branch"] = df.branch.str.strip()

    return df


def _read_hitran_file(fname, columns, count, linereturnformat):
    """
    Returns
    -------

    data: np.ndarray
    """

    dt = _create_dtype(columns, linereturnformat)

    return np.fromfile(fname, dtype=dt, count=count)


def _create_dtype(columns, linereturnformat):
    """Create dtype from specific columns.

    Returns
    -------
    dt: dtypes
    """
    # ... Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()] + [
        ("_linereturn", linereturnformat)
    ]
    # ... _linereturn is to capture the line return symbol. We delete it afterwards
    return _format_dtype(dtype)


def _format_dtype(dtype):
    """Format dtype from specific columns.

    Crash with hopefully helping error message
    """
    try:
        dt = np.dtype([(str(k), c) for k, c in dtype])
        # Note: dtype names cannot be `unicode` in Python2. Hence the str()
    except TypeError as err:
        # Cant read database. Try to be more explicit for user before crashing
        try:
            print("Data type")
            print("-" * 30)
            for (k, c) in dtype:
                print(str(k), "\t\t", c)
            print("-" * 30)
        finally:
            raise TypeError("Couldnt read datatype. See details above.") from err
    return dt


def _cast_to_dtype(data, dtype):
    """Cast array to certain type, crash with hopefull helping error message.

    Return casted data.

    Parameters
    ----------
    data: array to cast
    dtype: (ordered) list of (param, type)
    """
    dt = _format_dtype(dtype)

    try:
        data = np.array(data, dtype=dt)
    except ValueError as err:
        try:
            # Cant read database. Try to be more explicit for user before crashing
            # ... identify faulty row
            print("Cant cast data to specific dtype. Trying row by row:")
            for r in range(len(data)):
                try:
                    np.array(data[r], dtype=dt)
                except ValueError:
                    break
            print(f"Error may be on row {r}:")
            print("-" * 30)
            for i in range(len(data[r])):
                print(i, dtype[i], "\t\t", np.array(data[r][i], dtype=dt[i]))
            print("-" * 30)
            print(">>> Next param:", dtype[i], ". Value:", data[r][i], "\n")
        finally:
            raise ValueError(
                "Cant cast data to specific dtype. Tried column by column. See results above"
            ) from err

    return data


def drop_object_format_columns(df, verbose=True):
    """Remove 'object' columns in a pandas DataFrame.

    They are not useful to us at this time, and they slow down all
    operations (as they are converted to 'object' in pandas DataFrame).
    If you want to keep them, better convert them to some numeric values
    """

    objects = [k for k, v in df.dtypes.items() if v == object]
    for k in objects:
        del df[k]
    if verbose >= 2 and len(objects) > 0:
        print(
            (
                "The following columns had the `object` format and were removed: {0}".format(
                    objects
                )
            )
        )
    return df


def replace_PQR_with_m101(df):
    """Return P, Q, R in column ``branch`` with -1, 0, 1 to get a fully numeric
    database. This improves performances quite a lot, as Pandas doesnt have a
    fixed-string dtype hence would use the slow ``object`` dtype.

    Parameters
    ----------

    df: pandas Dataframe
        ``branch`` must be a column name

    Returns
    -------

    None:
        ``df`` is is modified in place
    """

    # Note: somehow pandas updates dtype automatically. We have to check
    # We just have to replace the column:
    if df.dtypes["branch"] != np.int64:
        new_col = df["branch"].replace(["P", "Q", "R"], [-1, 0, 1])
        df["branch"] = new_col

    assert df.dtypes["branch"] == np.int64



# -*- coding: utf-8 -*-
"""

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests

>>> pytest -m fast

"""

from os.path import basename

import matplotlib.pyplot as plt
import numpy as np
import pytest

from radis.misc.printer import printm
from radis.test.utils import setup_test_line_databases

fig_prefix = basename(__file__) + ": "

# %% Test routines


@pytest.mark.fast
def test_populations(verbose=True, *args, **kwargs):
    """Test that vib and rovib populations are calculated correctly"""

    from radis.lbl import SpectrumFactory
    from radis.misc.basics import all_in

    export = ["vib", "rovib"]

    sf = SpectrumFactory(
        2000,
        2300,
        export_populations=export,
        cutoff=1e-25,
        isotope="1",
    )
    sf.warnings.update(
        {"MissingSelfBroadeningWarning": "ignore", "VoigtBroadeningWarning": "ignore"}
    )
    sf.load_databank("HITRAN-CO-TEST")
    sf.misc.export_rovib_fraction = True
    # we test that "tabulation" and "export_population" are incompatible
    sf.params.parsum_mode = "tabulation"
    with pytest.raises(ValueError) as err:
        s = sf.non_eq_spectrum(2000, 2000)
    assert (
        str(err.value)
        == "Cannot update populations of individual levels with `tabulation` mode. Choose `update_populations=False` or `mode='full summation'`"
    )

    sf.params.parsum_mode = "full summation"  # won't be default at some point
    s = sf.non_eq_spectrum(2000, 2000)

    pops = sf.get_populations(export)
    if not all_in(["rovib", "vib"], list(pops["CO"][1]["X"].keys())):
        raise AssertionError(
            "vib and rovib levels should be defined after non_eq_spectrum calculation!"
        )
    if not "nvib" in list(pops["CO"][1]["X"]["vib"].keys()):
        raise AssertionError(
            "Vib populations should be defined after non_eq_spectrum calculation!"
        )

    s = sf.eq_spectrum(300)

    pops = sf.get_populations(export)
    if "nvib" in list(pops["CO"][1]["X"]["vib"].keys()):
        raise AssertionError(
            "Vib levels should not be defined anymore after eq_spectrum calculation!"
        )

    # Any of these is True and something went wrong
    s = sf.non_eq_spectrum(2000, 2000)
    s2 = sf.non_eq_spectrum(300, 300)

    # printm(all(s2.get_vib_levels(isotope=1) == s.get_vib_levels(isotope=1)))
    assert not (s2.get_vib_levels() is s.get_vib_levels())
    assert not (s2.get_rovib_levels() == s.get_rovib_levels()).all().all()
    assert not (s2.get_rovib_levels() is s.get_rovib_levels())

    return True  # if no AssertionError


@pytest.mark.fast
def test_rescaling_path_length(
    debug=False, plot=False, verbose=True, warnings=True, *args, **kwargs
):
    """Test rescaling functions"""

    if plot:  # Make sure matplotlib is interactive so that test are not stuck
        plt.ion()

    from radis.lbl import SpectrumFactory

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    Tgas = 1500
    sf = SpectrumFactory(
        wavelength_min=4400,
        wavelength_max=4800,
        mole_fraction=0.01,
        #                         path_length=0.1,
        cutoff=1e-25,
        wstep=0.005,
        isotope=[1],
        self_absorption=True,
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    #        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
    sf.load_databank("HITRAN-CO-TEST")
    s1 = sf.non_eq_spectrum(Tgas, Tgas, path_length=0.01)
    s2 = sf.non_eq_spectrum(Tgas, Tgas, path_length=3)
    s1.rescale_path_length(3)

    if plot:
        fig = plt.figure(fig_prefix + "Rescaling path length")
        s2.plot("radiance_noslit", nfig=fig.number, lw=3, label="L=3m")
        s1.plot(
            "radiance_noslit",
            nfig=fig.number,
            color="r",
            label="L=0.01m, rescaled to 3m",
        )
        plt.title("Non optically thin rescaling")
        plt.legend()
        plt.tight_layout()

    if verbose:
        printm("Test rescaling:")
        printm(
            "... Difference: {0:.2f}%".format(
                abs(s1.get_power() / s2.get_power() - 1) * 100
            )
        )

    assert np.isclose(s2.get_power(), s1.get_power(), 2e-3)


@pytest.mark.fast
def test_rescaling_mole_fraction(
    debug=False, plot=False, verbose=True, warnings=True, *args, **kwargs
):
    """Test rescaling functions"""

    from radis.lbl import SpectrumFactory

    if plot:  # Make sure matplotlib is interactive so that test are not stuck
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    Tgas = 1500
    sf = SpectrumFactory(
        wavelength_min=4400,
        wavelength_max=4800,
        #                     mole_fraction=1,
        path_length=0.1,
        mole_fraction=0.01,
        cutoff=1e-25,
        wstep=0.005,
        isotope=[1],
        self_absorption=True,
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.load_databank("HITRAN-CO-TEST")
    error = []
    N = [1e-3, 1e-2, 1e-1, 0.3, 0.6, 1]  # first is ref
    for Ni in N:
        s1 = sf.non_eq_spectrum(Tgas, Tgas, mole_fraction=N[0])
        sN = sf.non_eq_spectrum(Tgas, Tgas, mole_fraction=Ni)
        s1.rescale_mole_fraction(Ni)
        error.append(sN.get_power() / s1.get_power())

    if plot:
        plt.figure(fig_prefix + "Rescaling mole fractions")
        plt.plot(N, error, "-ok")
        plt.scatter(
            N[0],
            error[0],
            s=200,
            facecolors="none",
            edgecolors="r",
            label="reference",
        )
        plt.xlabel("Mole fraction")
        plt.ylabel("scaled energy / ab initio energy")
        plt.xscale("log")
        plt.legend()
        plt.title("Effect of scaling mole fraction w/o lineshape update")
        plt.tight_layout()

    # less than 1% error when rescaling from 1e-3 to 0.6
    assert abs(error[-2] - 1) < 0.01


@pytest.mark.fast
def test_medium(plot=False, verbose=True, debug=False, warnings=True, *args, **kwargs):
    """Test effect of propagating medium"""

    from radis.lbl.factory import SpectrumFactory

    if plot:  # Make sure matplotlib is interactive so that test are not stuck
        plt.ion()

    T = 300

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    pl = SpectrumFactory(
        wavenum_min=2171.5,
        wavenum_max=2174,
        mole_fraction=0.01,
        medium="vacuum",
        isotope="1,2",
    )
    pl.warnings["MissingSelfBroadeningWarning"] = "ignore"
    pl.load_databank("HITRAN-CO-TEST")
    s = pl.non_eq_spectrum(Tvib=T, Trot=T)  # , Ttrans=300)

    pla = SpectrumFactory(
        wavenum_min=2171.5,
        wavenum_max=2174,
        mole_fraction=0.01,
        medium="air",
        isotope="1,2",
    )
    pla.load_databank("HITRAN-CO-TEST")
    s_air = pla.non_eq_spectrum(Tvib=T, Trot=T)  # , Ttrans=300)

    if plot:
        plt.figure(fig_prefix + "Propagating medium conversions")
        s.plot(wunit="nm_vac", nfig="same", lw=3, label="vacuum")
        s.plot(wunit="nm", nfig="same", label="air")

    assert np.allclose(s.get_wavenumber(), s_air.get_wavenumber())
    assert np.allclose(
        s.get_wavelength(medium="vacuum"), s_air.get_wavelength(medium="vacuum")
    )
    assert np.allclose(
        s.get_wavelength(medium="air"), s_air.get_wavelength(medium="air")
    )
    assert all(s.get_wavelength(medium="vacuum") > s.get_wavelength(medium="air"))


def _run_testcases(
    plot=False, verbose=True, debug=False, warnings=True, *args, **kwargs
):
    """Test procedures

    Parameters
    ----------

    debug: boolean
        swamps the console namespace with local variables. Default ``False``

    """

    # Test populations
    # ----------
    test_populations(verbose=verbose, *args, **kwargs)

    # Test conversion of intensity cm-1 works
    # -------------
    #    test_intensity_conversion(debug=debug, verbose=verbose, *args, **kwargs)   # Moved in RADIS

    # Test updating / rescaling functions (no self absorption)
    # ---------
    #    test_rescaling_function(debug=debug, *args, **kwargs)     # Moved in RADIS
    test_rescaling_path_length(
        plot=plot, verbose=verbose, debug=debug, warnings=warnings, *args, **kwargs
    )
    test_rescaling_mole_fraction(
        plot=plot, verbose=verbose, debug=debug, warnings=warnings, *args, **kwargs
    )

    # Test propagating medium
    test_medium(
        plot=plot, verbose=verbose, debug=debug, warnings=warnings, *args, **kwargs
    )

    return True


if __name__ == "__main__":
    printm("Test spectrum: ", _run_testcases(debug=False, plot=True))



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 26 20:36:38 2021

@author: erwan
"""

import pytest

from radis.io.hdf5 import HDF5Manager, hdf2df
from radis.io.hitemp import fetch_hitemp
from radis.misc.config import getDatabankEntries


@pytest.mark.fast
def test_hdf5_io_engines(*args, **kwargs):
    """Test different engines implemented in :py:class:`radis.io.hdf5.HDF5Manager`"""

    import os
    from os.path import exists

    for file in ["test_pytables.h5", "test_h5py.h5"]:
        if exists(file):
            os.remove(file)

    # Test data
    import h5py
    import numpy as np
    import pandas as pd

    df0 = pd.DataFrame({"a": np.arange(10) ** 2, "b": np.arange(10) ** 3})
    metadata0 = {"some_metadata": True}
    # ... a Pandas HDFStore file
    df0.to_hdf("test_pytables.h5", "df")
    # ... a h5py file with same content
    with h5py.File("test_h5py.h5", "w") as f:
        # group = f.create_group("df")
        for c in df0.columns:
            f.create_dataset(c, data=df0[c])

    # Test Pytables engine : add_metadata ; load ; read_metadata
    manager = HDF5Manager(engine="pytables")
    manager.add_metadata("test_pytables.h5", metadata0)

    df = manager.load("test_pytables.h5")
    assert (df == df0).all().all()
    assert manager.read_metadata("test_pytables.h5") == metadata0

    # Test h5py engine : add_metadata ; load ; read_metadata
    manager = HDF5Manager(engine="h5py")
    manager.add_metadata("test_h5py.h5", metadata0)
    df = manager.load("test_h5py.h5", key=None)
    assert (df == df0).all().all()
    assert manager.read_metadata("test_h5py.h5") == metadata0

    # Test vaex engine : add_metadata ; load ; read_
    # .. also able to read h5py files
    manager = HDF5Manager(engine="vaex")
    manager.add_metadata("test_h5py.h5", metadata0)
    df = manager.load("test_h5py.h5", key=None)
    # this time; df is a vaex DataFrame
    # ... it keeps the file open so we should close the file handle
    assert (df.to_pandas_df() == df0).all().all()
    df.close()
    assert manager.read_metadata("test_h5py.h5") == metadata0


@pytest.mark.needs_connection
def test_local_hdf5_lines_loading(*args, **kwargs):
    """
    We use the OH HITEMP line database to test :py:func:`~radis.io.hitemp.fetch_hitemp`
    and :py:func:`~radis.io.hdf5.hdf2df`

    - Partial loading (only specific wavenumbers)
    - Only certain isotopes
    - Only certain columns

    """

    fetch_hitemp("OH")  # to initialize the database

    path = getDatabankEntries("HITEMP-OH")["path"]

    # Initialize the database
    fetch_hitemp("OH")
    path = getDatabankEntries("HITEMP-OH")["path"][0]
    df = hdf2df(path)
    wmin, wmax = df.wav.min(), df.wav.max()
    assert wmin < 2300  # needed for next test to be valid
    assert wmax > 2500  # needed for next test to be valid
    assert len(df.columns) > 5  # many columns loaded by default
    assert len(df.iso.unique()) > 1

    # Test loading only certain columns
    df = hdf2df(path, columns=["wav", "int"])
    assert len(df.columns) == 2 and "wav" in df.columns and "int" in df.columns

    # Test loading only certain isotopes
    df = hdf2df(path, isotope="2")
    assert df.iso.unique() == 2

    # Test partial loading of wavenumbers
    df = hdf2df(path, load_wavenum_min=2300, load_wavenum_max=2500)
    assert df.wav.min() >= 2300
    assert df.wav.max() <= 2500

    # Test with only one
    assert hdf2df(path, load_wavenum_min=2300).wav.min() >= 2300

    # Test with the other
    assert hdf2df(path, load_wavenum_max=2500).wav.max() <= 2500


if __name__ == "__main__":
    test_hdf5_io_engines()
    test_local_hdf5_lines_loading()



# -*- coding: utf-8 -*-
"""Parser for CDSD-HITEMP, CDSD-4000 format.

Routine Listing
---------------

- :func:`~radis.io.cdsd.cdsd2df`


References
----------

CDSD-4000 manual

-------------------------------------------------------------------------------
"""

# TODO: remove wangl2  after loading database (wangl is enough, with 1=e and 2=f)


import time
from collections import OrderedDict
from os.path import exists, getmtime

import radis
from radis.io.cache_files import cache_file_name, load_h5_cache_file, save_to_hdf
from radis.io.tools import (
    drop_object_format_columns,
    parse_hitran_file,
    replace_PQR_with_m101,
)

# fmt: off
columns_cdsdhitemp = OrderedDict(
    [
        # name    # format # type  # description                                 # unit
        ("id", ("a2", int, "Molecular number", "")),
        ("iso", ("a1", int, "isotope number", "")),
        ("wav", ("a12", float, "vacuum wavenumber", "cm-1")),
        ("int", ("a10", float, "intensity at 296K", "cm-1/(molecule/cm-2)")),
        ("A", ("a10", float, "Einstein A coefficient", "s-1")),
        ("airbrd", ("a5", float, "air-broadened half-width at 296K", "cm-1.atm-1")),
        ("selbrd", ("a5", float, "self-broadened half-width at 296K", "cm-1.atm-1")),
        ("El", ("a10", float, "lower-state energy", "cm-1")),
        ("Tdpair", ("a4", float, "temperature-dependance exponent for Gamma air", "")),
        ("Pshft",("a8", float, "air pressure-induced line shift at 296K", "cm-1.atm-1")),
        # skip 1 columns   . WARNING. They didn't say that in the doc. But it doesn't make sense if i don't
        ("Tdpsel", ("a5", float, "temperature dependance exponent for gamma self", "")),
        ("v1u", ("a3", int, "upper state vibrational number v1", "")),
        ("v2u", ("a2", int, "upper state vibrational number v2", "")),
        ("l2u", ("a2", int, "upper state vibrational number l2", "")),
        ("v3u", ("a2", int, "upper state vibrational number v3", "")),
        ("ru", ("a1", int, "upper state vibrational number r", "")),
        # skip 5 columns (v1l format becomes 3+5)
        ("v1l", ("a8", int, "lower state vibrational number v1", "")),
        ("v2l", ("a2", int, "lower state vibrational number v2", "")),
        ("l2l", ("a2", int, "lower state vibrational number l2", "")),
        ("v3l", ("a2", int, "lower state vibrational number v3", "")),
        ("rl", ("a1", int, "lower state vibrational number r", "")),
        ("polyu", ("a3", int, "upper state polyad", "")),
        ("wangu", ("a2", int, "upper state Wang symmetry", "")),
        ("ranku", ("a4", int, "upper state ranking number", "")),
        ("polyl", ("a3", int, "lower state polyad", "")),
        ("wangl", ("a2", int, "lower state Wang symmetry", "")),
        ("rankl", ("a4", int, "lower state ranking number", "")),
        # skip 2 columns (PQR format becomes 1+2)
        ("branch", ("a3", str, "O, P, Q, R, S branch symbol", "")),
        ("jl", ("a3", int, "lower state rotational quantum number", "")),
        ("wangl2", ("a1", str, "lower state Wang symmetry", "(e or f)")),
        ("lsrc", ("a5", int, "line source", "")),
    ]
)

columns_4000 = OrderedDict(
    [
        # name    # format # type  # description                                 # unit
        ("id", ("a2", int, "Molecular number", "")),
        ("iso", ("a1", int, "isotope number", "")),
        ("wav", ("a12", float, "vacuum wavenumber", "cm-1")),
        ("int", ("a10", float, "intensity at 296K", "cm-1/(molecule/cm-2)")),
        ("A", ("a10", float, "Einstein A coefficient", "s-1")),
        ("airbrd", ("a5", float, "air-broadened half-width at 296K", "cm-1.atm-1")),
        ("selbrd", ("a5", float, "self-broadened half-width at 296K", "cm-1.atm-1")),
        ("El", ("a10", float, "lower-state energy", "cm-1")),
        ("Tdpair", ("a4", float, "temperature-dependance exponent for Gamma air", "")),
        ("Pshft",("a8", float, "air pressure-induced line shift at 296K", "cm-1.atm-1")),
        # skip 1 columns  (Tdpsel becomes 4+1 = 5)
        ("Tdpsel", ("a5", float, "temperature dependance exponent for gamma self", "")),
        ("v1u", ("a3", int, "upper state vibrational number v1", "")),
        ("v2u", ("a2", int, "upper state vibrational number v2", "")),
        ("l2u", ("a2", int, "upper state vibrational number l2", "")),
        ("v3u", ("a2", int, "upper state vibrational number v3", "")),
        ("ru", ("a2", int, "upper state vibrational number r", "")),
        # skip 5 columns (v1l format becomes 3+3=6)
        ("v1l", ("a6", int, "lower state vibrational number v1", "")),
        ("v2l", ("a2", int, "lower state vibrational number v2", "")),
        ("l2l", ("a2", int, "lower state vibrational number l2", "")),
        ("v3l", ("a2", int, "lower state vibrational number v3", "")),
        ("rl", ("a2", int, "lower state vibrational number r", "")),
        ("polyu", ("a3", int, "upper state polyad", "")),
        ("wangu", ("a2", int, "upper state Wang symmetry", "")),
        ("ranku", ("a4", int, "upper state ranking number", "")),
        ("polyl", ("a3", int, "lower state polyad", "")),
        ("wangl", ("a2", int, "lower state Wang symmetry", "")),
        ("rankl", ("a4", int, "lower state ranking number", "")),
        # skip 2 columns (PQR format becomes 1+2)
        ("branch", ("a3", str, "O, P, Q, R, S branch symbol", "")),
        ("jl", ("a3", int, "lower state rotational quantum number", "")),
        ("wangl2", ("a1", str, "lower state Wang symmetry", "(e or f)")),
    ]
)
# fmt: on


def cdsd2df(
    fname,
    version="hitemp",
    cache=True,
    verbose=True,
    drop_non_numeric=True,
    load_wavenum_min=None,
    load_wavenum_max=None,
):
    """Convert a CDSD-HITEMP [1]_ or CDSD-4000 [2]_ file to a Pandas dataframe.

    Parameter
    ----------
    fname: str
        CDSD file name
    version: str ('4000', 'hitemp')
        CDSD version
    cache: boolean, or 'regen'
        if ``True``, a pandas-readable HDF5 file is generated on first access,
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not
        taken into account). If ``False``, no database is used. If 'regen', temp
        file are reconstructed. Default ``True``.

    Other Parameters
    ----------------
    drop_non_numeric: boolean
        if ``True``, non numeric columns are dropped. This improves performances,
        but make sure all the columns you need are converted to numeric formats
        before hand. Default ``True``. Note that if a cache file is loaded it
        will be left untouched.
    load_wavenum_min, load_wavenum_max: float
        if not ``'None'``, only load the cached file if it contains data for
        wavenumbers above/below the specified value. See :py:func`~radis.io.cache_files.load_h5_cache_file`.
        Default ``'None'``.

    Returns
    -------
    df: pandas Dataframe
        dataframe containing all lines and parameters

    Notes
    -----

    CDSD-4000 Database can be downloaded from [3]_

    Performances: I had huge performance trouble with this function, because the files are
    huge (500k lines) and the format is to special (no space between numbers...)
    to apply optimized methods such as pandas's. A line by line reading isn't
    so bad, using struct to parse each line. However, we waste typing determining
    what every line is. I ended up using the fromfiles functions from numpy,
    not considering *\\n* (line return) as a special character anymore, and a second call
    to numpy to cast the correct format. That ended up being twice as fast.

        - initial:                      20s / loop
        - with mmap:                    worse
        - w/o readline().rstrip('\\n'):  still 20s
        - numpy fromfiles:              17s
        - no more readline, 2x fromfile 9s

    Think about using cache mode too:

        - no cache mode                 9s
        - cache mode, first time        22s
        - cache mode, then              2s

    Moving to HDF5:

    On cdsd_02069_02070 (56 Mb)

    Reading::

        cdsd2df(): 9.29 s
        cdsd2df(cache=True [old .txt version]): 2.3s
        cdsd2df(cache=True [new h5 version, table]): 910ms
        cdsd2df(cache=True [new h5 version, fixed]): 125ms

    Storage::

        %timeit df.to_hdf("cdsd_02069_02070.h5", "df", format="fixed")  337ms
        %timeit df.to_hdf("cdsd_02069_02070.h5", "df", format="table")  1.03s

    References
    ----------

    Note that CDSD-HITEMP is used as the line database for CO2 in HITEMP 2010

    .. [1] `HITEMP 2010, Rothman et al., 2010 <https://www.sciencedirect.com/science/article/pii/S002240731000169X>`_

    .. [2] `CDSD-4000 article, Tashkun et al., 2011 <https://www.sciencedirect.com/science/article/pii/S0022407311001154>`_

    .. [3] `CDSD-4000 database <ftp://ftp.iao.ru/pub/CDSD-4000/>`_

    See Also
    --------

    :func:`~radis.io.hitran.hit2df`
    """
    metadata = {}
    metadata["last_modification"] = time.ctime(getmtime(fname))
    if load_wavenum_min and load_wavenum_max:
        assert load_wavenum_min < load_wavenum_max

    if verbose >= 2:
        print(
            "Opening file {0} (format=CDSD {1}, cache={2})".format(
                fname, version, cache
            )
        )
        print("Last Modification time: {0}".format(metadata["last_modification"]))

    if version == "hitemp":
        columns = columns_cdsdhitemp
    elif version == "4000":
        columns = columns_4000
    else:
        raise ValueError("Unknown CDSD version: {0}".format(version))

    # Use cache file if possible
    fcache = cache_file_name(fname)
    if cache and exists(fcache):
        relevant_if_metadata_above = (
            {"wavenum_max": load_wavenum_min} if load_wavenum_min else {}
        )  # not relevant if wavenum_max of file is < wavenum min required
        relevant_if_metadata_below = (
            {"wavenum_min": load_wavenum_max} if load_wavenum_max else {}
        )  # not relevant if wavenum_min of file is > wavenum max required
        df = load_h5_cache_file(
            fcache,
            cache,
            valid_if_metadata_is=metadata,
            relevant_if_metadata_above=relevant_if_metadata_above,
            relevant_if_metadata_below=relevant_if_metadata_below,
            current_version=radis.__version__,
            last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
            verbose=verbose,
        )
        if df is not None:
            return df

    # %% Start reading the full file

    df = parse_hitran_file(fname, columns)

    # Remove non numerical attributes
    if drop_non_numeric:
        replace_PQR_with_m101(df)
        df = drop_object_format_columns(df, verbose=verbose)

    # cached file mode but cached file doesn't exist yet (else we had returned)
    if cache:
        new_metadata = {
            # Last modification time of the original file :
            "last_modification": time.ctime(getmtime(fname)),
            "wavenum_min": df.wav.min(),
            "wavenum_max": df.wav.max(),
        }
        if verbose:
            print(
                "Generating cache file {0} with metadata :\n{1}".format(
                    fcache, new_metadata
                )
            )
        try:
            save_to_hdf(
                df,
                fcache,
                metadata=new_metadata,
                version=radis.__version__,
                key="df",
                overwrite=True,
                verbose=verbose,
            )
        except PermissionError:
            if verbose:
                print("An error occured in cache file generation. Lookup access rights")
            pass

    # TODO : get only wavenum above/below 'load_only_wavenum_above', 'load_only_wavenum_below'
    # by parsing df.wav.   Completely irrelevant files are discarded in 'load_h5_cache_file'
    # but files that have partly relevant lines are fully loaded.
    # Note : cache file is generated with the full line list.

    return df


if __name__ == "__main__":
    from radis.test.test_io import test_hitemp

    print("Testing cdsd: ", test_hitemp())



# -*- coding: utf-8 -*-
"""Core of the line-by-line calculations
"""

# prevent cyclic imports:
from . import bands, base, broadening, calc, factory, labels, loader, overp
from .calc import calc_spectrum
from .factory import SpectrumFactory
from .overp import LevelsList

__all__ = ["LevelsList", "SpectrumFactory", "calc_spectrum"]



"""
`Profiler` class, for printing time spent on various steps during Spectrum
calculation under :py:class:`~radis.lbl.factory.SpectrumFactory` based on verbose level.

Also stores Spectrum calculation time dependent parameters, under the attribute
    :py:attr:`~radis.misc.profiler.Profiler.final`

Routine Listing
---------------

- :meth:`~radis.misc.profiler.Profiler.add_entry`
- :meth:`~radis.misc.profiler.Profiler.start`
- :meth:`~radis.misc.profiler.Profiler.add_time`
- :meth:`~radis.misc.profiler.Profiler.stop`
- :meth:`~radis.misc.profiler.Profiler._print`

-------------------------------------------------------------------------------
"""

from collections import OrderedDict
from time import time

from radis.misc.printer import printg


class Profiler(object):
    """A class to store Spectrum calculation time dependent parameters, under the attribute
    :py:attr:`~radis.misc.profiler.Profiler.final` of
    :py:class:`~radis.lbl.factory.SpectrumFactory`.

    It also hold functions to print all the entities based on verbose value.

    See Also
    --------

    :py:attr:`~radis.lbl.loader.DatabankLoader.input`,
    :py:attr:`~radis.misc.profiler.Profiler,

    """

    def __init__(self, verbose):
        super(Profiler, self).__init__()

        # Dev: Init here to be found by autocomplete
        self.initial = {}
        self.verbose = verbose
        self.final = OrderedDict()
        self.relative_time_percentage = {}

    # Creates profiler dictionary structure
    def add_entry(self, dictionary, key, verbose, count):
        if count == verbose:
            if key in dictionary:
                return
            else:
                dictionary[key] = {}

            return

        self.add_entry(dictionary[list(dictionary)[-1]], key, verbose, count + 1)

    def start(self, key, verbose_level, optional=""):
        if __debug__:
            self.initial[key] = {
                "start_time": time(),
                "verbose_level": verbose_level,
            }

            self.add_entry(self.final, key, verbose_level, 1)

        if len(optional) != 0 and self.verbose >= verbose_level:
            print(optional)

    # Adds time calculated for each key in profiler
    def add_time(self, dictionary, key, verbose, count, time_calculated):
        if count == verbose:
            executed_more_than_once = 0  # Checks if input key was executed before

            # Checking if value of key is empty or float
            if isinstance(dictionary[key], float):
                time_calculated = dictionary[key] + time_calculated
                executed_more_than_once = 1
            else:
                try:
                    if "value" in dictionary[key] and isinstance(
                        dictionary[key]["value"], float
                    ):
                        time_calculated = dictionary[key]["value"] + time_calculated
                        executed_more_than_once = 2
                except KeyError:
                    pass

            # Adding values
            if executed_more_than_once == 2:
                dictionary[key]["value"] = time_calculated
            elif executed_more_than_once == 1:
                dictionary[key] = time_calculated
            else:
                if len(dictionary[key]) != 0:
                    dictionary[key].update({"value": time_calculated})
                else:
                    dictionary[key] = time_calculated
            return

        self.add_time(
            dictionary[list(dictionary)[-1]], key, verbose, count + 1, time_calculated
        )

    def stop(self, key, details):
        if __debug__:
            items = self.initial.pop(key)
            time_calculated = time() - items["start_time"]

            if items["verbose_level"] == 1:
                # Profiler ends; Deserializing to Dictionary format
                self.final = dict(self.final)

            self.add_time(self.final, key, items["verbose_level"], 1, time_calculated)

            if self.verbose >= items["verbose_level"]:
                self._print(
                    items["verbose_level"],
                    details,
                    time_calculated=time_calculated,
                )

    def _print(self, verbose_level, details, time_calculated):

        if verbose_level == 1:
            print("{0:.2f}s -".format(time_calculated), details)
        elif verbose_level >= 2:
            printg(
                "..." * (verbose_level - 1),
                "{0:.2f}s -".format(time_calculated),
                details,
            )



# -*- coding: utf-8 -*-
"""

Notes
-----

Planck functions:
- planck: planck radiation with wavelength input
- planck_wn: planck radiation with wavenumber input
- sPlanck: a RADIS :class:`~radis.spectrum.spectrum.Spectrum` blackbody object

Example
-------

Generate Earth blackbody::

    s = sPlanck(wavelength_min=3000, wavelength_max=50000,
                T=288, eps=1)
    s.plot()

-------------------------------------------------------------------------------


"""

from numpy import arange, exp, inf, ones_like, zeros_like

from radis.phys.air import air2vacuum
from radis.phys.constants import c, c_CGS, h, h_CGS, k_b, k_b_CGS
from radis.phys.units import Unit as Q_
from radis.phys.units import conv2


def planck(lmbda, T, eps=1, unit="mW/sr/cm2/nm"):
    r"""Planck function for blackbody radiation.

    .. math::
        \epsilon \frac{2h c^2}{{\lambda}^5} \frac{1}{\operatorname{exp}\left(\frac{h c}{\lambda k T}\right)-1}

    Parameters
    ----------
    λ: np.array   (nm)
       wavelength
    T: float    (K)
        equilibrium temperature
    eps: grey-body emissivity
        default 1
    unit: output unit
        default 'mW/sr/cm2/nm'

    Returns
    -------
    np.array :  (mW.sr-1.cm-2/nm)
        equilibrium radiance

    See Also
    --------
    :py:func:`~radis.tools.blackbody.sPlanck`, :py:func:`~radis.phys.blackbody.planck_wn`
    """
    k = k_b
    lbd = lmbda * 1e-9
    iplanck = (
        eps * (2 * h * c ** 2 / lbd ** 5) * (1 / (exp(h * c / (lbd * k * T)) - 1))
    )  # S.I  (W.sr-1.m-3)
    iplanck *= 1e-10  # W.sr-1.m-3 >>> mW.sr-1.cm-2.nm-1

    if Q_(unit) != Q_("mW/sr/cm2/nm"):
        iplanck = conv2(iplanck, "mW/sr/cm2/nm", unit)

    return iplanck


def planck_wn(wavenum, T, eps=1, unit="mW/sr/cm2/cm-1"):
    r"""Planck function for blackbody radiation, wavenumber version.

    .. math::
        \epsilon 2h c^2 {\nu}^3 \frac{1}{\operatorname{exp}\left(\frac{h c \nu}{k T}\right)-1}

    Parameters
    ----------
    wavenum: np.array   (cm-1)
       wavenumber
    T: float    (K)
        equilibrium temperature
    eps: grey-body emissivity
        default 1
    unit: str
        output unit. Default 'mW/sr/cm2/cm-1'


    Returns
    -------
    np.array :  default (mW/sr/cm2/cm-1)
        equilibrium radiance

    See Also
    --------
    :py:func:`~radis.tools.blackbody.sPlanck`, :py:func:`~radis.phys.blackbody.planck`
    """
    k = k_b_CGS
    h = h_CGS
    c = c_CGS

    iplanck = (
        eps
        * (2 * h * c ** 2 * wavenum ** 3)
        * (1 / (exp(h * c * wavenum / (k * T)) - 1))
    )
    # iplanck in erg/s/sr/cm2/cm-1
    iplanck *= 1e-4  # erg/s/sr/cm2/cm-1 > mW/sr/cm^2/cm-1

    if Q_(unit) != Q_("mW/sr/cm2/cm-1"):
        iplanck = conv2(iplanck, "mW/sr/cm2/cm-1", unit)

    return iplanck


# %% Predefined Spectra objects


def sPlanck(
    wavenum_min=None,
    wavenum_max=None,
    wavelength_min=None,
    wavelength_max=None,
    T=None,
    eps=1,
    wstep=0.01,
    medium="air",
    **kwargs
):
    r"""Return a RADIS :py:class:`~radis.spectrum.spectrum.Spectrum` object with blackbody radiation.

    It's easier to plug in a :py:func:`~radis.los.slabs.SerialSlabs` line-of-sight than the Planck
    radiance calculated by :py:func:`~radis.phys.blackbody.planck`.
    And you don't need to worry about units as they are handled internally.

    See :py:class:`~radis.spectrum.spectrum.Spectrum` documentation for more information

    Parameters
    ----------
    wavenum_min / wavenum_max: ():math:`cm^{-1}`)
        minimum / maximum wavenumber to be processed in :math:`cm^{-1}`.
    wavelength_min / wavelength_max: (:math:`nm`)
        minimum / maximum wavelength to be processed in :math:`nm`.
    T: float (K)
        blackbody temperature
    eps: float [0-1]
        blackbody emissivity. Default ``1``

    Other Parameters
    ----------------
    wstep: float (cm-1 or nm)
        wavespace step for calculation
    **kwargs: other keyword inputs
        all are forwarded to spectrum conditions. For instance you can add
        a 'path_length=1' after all the other arguments

    Examples
    --------
    Generate Earth blackbody::

        s = sPlanck(wavelength_min=3000, wavelength_max=50000,
                    T=288, eps=1)
        s.plot()

    Examples using sPlanck :

    .. minigallery:: radis.sPlanck

    References
    ----------
    In wavelength:

    .. math::
        \epsilon \frac{2h c^2}{{\lambda}^5} \frac{1}{\operatorname{exp}\left(\frac{h c}{\lambda k T}\right)-1}

    In wavenumber:

    .. math::
        \epsilon 2h c^2 {\nu}^3 \frac{1}{\operatorname{exp}\left(\frac{h c \nu}{k T}\right)-1}

    See Also
    --------
    :py:func:`~radis.phys.blackbody.planck`, :py:func:`~radis.phys.blackbody.planck_wn`

    """
    from radis.spectrum.spectrum import Spectrum

    # Check inputs
    if (wavelength_min is not None or wavelength_max is not None) and (
        wavenum_min is not None or wavenum_max is not None
    ):
        raise ValueError("You cannot give both wavelengths and wavenumbers")

    if wavenum_min is not None and wavenum_max is not None:
        assert wavenum_min < wavenum_max
        waveunit = "cm-1"
    else:
        assert wavelength_min < wavelength_max
        if medium == "air":
            waveunit = "nm"
        elif medium == "vacuum":
            waveunit = "nm_vac"
        else:
            raise ValueError(medium)

    if T is None:
        raise ValueError("T must be defined")

    if not (eps >= 0 and eps <= 1):
        raise ValueError("Emissivity must be in [0-1]")

    # Test range is correct:
    if waveunit == "cm-1":
        # generate the vector of wavenumbers (shape M)
        w = arange(wavenum_min, wavenum_max + wstep, wstep)
        Iunit = "mW/sr/cm2/cm-1"
        I = planck_wn(w, T, eps=eps, unit=Iunit)
    else:
        # generate the vector of lengths (shape M)
        w = arange(wavelength_min, wavelength_max + wstep, wstep)
        Iunit = "mW/sr/cm2/nm"
        if waveunit == "nm_vac":
            w_vac = w
        elif waveunit == "nm":
            w_vac = air2vacuum(w)
        # calculate planck with wavelengths in vacuum
        I = planck(w_vac, T, eps=eps, unit=Iunit)

    conditions = {"wstep": wstep}
    # add all extra parameters in conditions (ex: path_length)
    conditions.update(**kwargs)

    return Spectrum(
        quantities={
            "radiance_noslit": (w, I),
            "transmittance_noslit": (w, zeros_like(w)),
            "absorbance": (w, ones_like(w) * inf),
        },
        conditions=conditions,
        units={"radiance_noslit": Iunit, "transmittance_noslit": "", "absorbance": ""},
        cond_units={"wstep": waveunit},
        wunit=waveunit,
        name="Planck {0}K, eps={1:.2g}".format(T, eps),
    )


if __name__ == "__main__":
    from radis.test.phys.test_blackbody import _run_testcases

    _run_testcases(plot=True)
